from ..util import prettify


class Group:
    """
    This object isn't returned by any endpoints yet, it is here purely as a reference for :class:`UserGroup`

    **Attributes**

    id: :class:`int`

    identifier: :class:`str`
        Unique string to identify the group.

    is_probationary: :class:`str`
        Whether members of this group are considered probationary.

    has_playmodes: :class:`bool`
        If this group associates GameModes with a user's membership, e.g. BN/NAT members

    has_listing: :class:`bool`
        Whether this group displays a listing at /groups/{id}

    name: :class:`str`

    short_name: :class:`str`
        Short name of the group for display.

    colour: :class:`str` or :class:`NoneType`

    **Optional Attributes**

    description: :class:`Description` or :class:`NoneType`
        A dictionary with keys html and markdown.
    """
    __slots__ = (
        "id", "identifier", "is_probationary", "has_playmodes", "has_listing",
        "name", "short_name", "colour", "description"
    )

    def __init__(self, data):
        self.id = data['id']
        self.identifier = data['identifier']
        self.is_probationary = data['is_probationary']
        self.has_playmodes = data['has_playmodes']
        self.name = data['name']
        self.short_name = data['short_name']
        self.colour = data['colour']
        self.description = Description(data['description']) if data.get('description') is not None else None

    def __repr__(self):
        return prettify(self, 'name')


class Description:
    """
    **Attributes**

    html: :class:`str`

    markdown: :class:`str`
    """

    def __init__(self, data):
        self.html = data['html']
        self.markdown = data['markdown']

    def __repr__(self):
        return prettify(self)


class UserGroup(Group):
    """
    Describes the :class:`Group` membership of a :class:`User`.
    It contains all of the attributes of the :class:`Group`, in addition to what is listed here.

    **Attributes**

    playmodes: Sequence[:class:`str`]
        GameModes associated with this membership (null if has_playmodes is unset).
    """
    __slots__ = (
        "playmodes",
    )

    def __init__(self, data):
        super().__init__(data)
        self.playmodes = data['playmodes']

    def __repr__(self):
        return super().__repr__()



from .user import CurrentUserAttributes, UserCompact
from dateutil import parser

from ..util import prettify
from ..enums import ChatChannelType


class ChatChannel:
    """
    Represents an individual chat "channel" in the game.

    **Attributes**

    Some attributes will be :class:`NoneType`

    channel_id: :class:`int`

    current_user_attributes: :class:`CurrentUserAttributes`
        only present on some responses

    name: :class:`str`

    description: :class:`str` or :class:`NoneType`

    icon: :class:`str`
        display icon for the channel

    type: :class:`str`
        Below are the channel types and their permission checks for joining/messaging:
        PUBLIC
            None
        PRIVATE
            is player in the allowed groups? (channel.allowed_groups)
        MULTIPLAYER
            is player currently in the mp game?
        SPECTATOR
            None
        TEMPORARY
            deprecated
        PM
            Is either user blocking the other? If so, deny.
            Does the target only accept PMs from friends? Is the current user a friend? If not, deny.
        GROUP
            is player in channel? (user_channels)

    first_message_id: :class:`int`
        message_id of first message (only returned in presence responses)

    last_message_id: :class:`int`
        message_id of last known message (only returned in presence responses)

    recent_messages: class:`list`
        list containing objects of type :class:`ChatMessage`. Up to 50 most recent messages.

    moderated: :class:`bool`
        user can't send message when the value is true (only returned in presence responses)

    users: :class:`list`
        list of user_id that are in the channel (not included for PUBLIC channels)
    """
    __slots__ = (
        "channel_id", "current_user_attributes", "name", "description", "icon",
        "type", "last_read_id", "last_message_id", "recent_messages",
        "moderated", "users"
    )

    def __init__(self, data):
        self.channel_id = data.get('channel_id')
        self.current_user_attributes = CurrentUserAttributes(data['current_user_attributes'],
                                                             "ChatChannelUserAttributes") \
            if data.get("current_user_attributes") is not None else None
        self.name = data.get("name")
        self.description = data.get("description")
        self.icon = data.get("icon")
        self.type = ChatChannelType(data["type"]) if data.get("type") is not None else None
        self.last_read_id = data.get("last_read_id")
        self.last_message_id = data.get("last_message_id")
        self.recent_messages = list(map(ChatMessage, data.get('recent_messages', [])))
        self.moderated = data.get("moderated")
        self.users = data.get("users")

    def __repr__(self):
        return prettify(self, 'name')


class ChatMessage:
    """
    Represents an individual Message within a :class:`ChatChannel`.

    **Attributes**

    message_id: :class:`int`
        unique identifier for message

    sender_id: :class:`int`
        user_id of the sender

    channel_id: :class:`int`
        channel_id of where the message was sent

    timestamp: :class:`datetime.datetime`
        when the message was sent

    content: :class:`str`
        message content

    is_action: :class:`bool`
        was this an action? i.e. /me dances

    sender: :class:`UserCompact`
        embeded :class:`UserCompact` object to save additional api lookups
    """
    __slots__ = (
        "message_id", "sender_id", "channel_id", "timestamp", "content", "is_action",
        "sender"
    )

    def __init__(self, data):
        self.message_id = data['message_id']
        self.sender_id = data['sender_id']
        self.channel_id = data['channel_id']
        self.timestamp = parser.parse(data['timestamp'])
        self.content = data['content']
        self.is_action = data['is_action']
        self.sender = UserCompact(data['sender'])

    def __repr__(self):
        return prettify(self, 'content')



from .auth import AuthHandler, LazerAuthHandler
from .client import Client
from .exceptions import *
from .notification import NotificationWebsocket
from .asyncio.client import AsynchronousClient
from .enums import *
from .objects import *
from .util import (
    BeatmapsetSearchFilter,
    PlaylistItemUtil,
    NotificationsUtil,
    IdentitiesUtil,
)


__version__ = '0.6.0'



from dateutil import parser

from ..util import prettify
from ..enums import NotificationCategory, ObjectType, NotificationType, ChatChannelType, GameModeStr


class Notification:
    """
    Represents a notification object.
    Go to :class:`Details` object to see details for each event

    **Attributes**

    id: :class:`int`

    name: :class:`NotificationType`
        The type of event

    created_at: :class:`datetime.datetime`

    object_type: :class:`ObjectType`

    object_id: :class:`int`

    source_user_id: :class:`int`

    is_read: :class:`bool`

    details: Optional[Union[
    :class:`BeatmapOwnerChangeDetails`, :class:`BeatmapsetDiscussionLockDetails`,
    :class:`BeatmapsetDiscussionPostDetails`, :class:`BeatmapsetDiscussionQualifiedProblemDetails`,
    :class:`BeatmapsetDiscussionReviewNewDetails`, :class:`BeatmapsetDiscussionUnlockDetails`,
    :class:`BeatmapsetDisqualifyDetails`, :class:`BeatmapsetLoveDetails`,
    :class:`BeatmapsetNominateDetails`, :class:`BeatmapsetQualifyDetails`,
    :class:`BeatmapsetRankDetails`, :class:`BeatmapsetRemoveFromLovedDetails`,
    :class:`BeatmapsetResetNominationsDetails`, :class:`ChannelAnnouncementDetails`,
    :class:`ChannelMessageDetails`, :class:`CommentNewDetails`,
    :class:`ForumTopicReplyDetails`, :class:`UserAchievementUnlockDetails`,
    :class:`UserBeatmapsetNewDetails`, :class:`UserBeatmapsetReviveDetails`
    ]]
        Details of the notification.

    source_user_id: Optional[int]


    **Documented event names and their attribute meanings**

    beatmapset_discussion_lock
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who locked discussion

    beatmapset_discussion_post_new
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: Poster of the discussion

    beatmapset_discussion_unlock
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who unlocked discussion

    beatmapset_disqualify
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who disqualified beatmapset

    beatmapset_love
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who promoted beatmapset to Loved

    beatmapset_nominate
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who nominated beatmapset

    beatmapset_qualify
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User whom beatmapset nomination triggered qualification

    beatmapset_remove_from_loved
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who removed beatmapset from Loved

    beatmapset_reset_nominations
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who locked discussion

    channel_message
        object_id: Channel id
        object_type: channel
        source_user_id: User who posted message

    forum_topic_reply
        object_id: Topic id
        object_type: forum_topic
        source_user_id: User who posted message
    """
    __slots__ = (
        "id", "name", "created_at", "object_type", "object_id", "is_read",
        "source_user_id", "details"
    )

    def __init__(self, data):
        self.id = data['id']
        self.name = NotificationType(data['name'])
        self.created_at = parser.parse(data['created_at'])
        self.object_type = ObjectType(data['object_type'])
        self.object_id = data['object_id']
        self.is_read = data['is_read']
        self.details = Details(data['details'], self.name)

        self.source_user_id = data.get('source_user_id', None)

    def __repr__(self):
        return prettify(self, 'name', 'details')


class ReadNotification:
    """
    Represents a read notification.

    **Attributes**

    category: :class:`str`

    id: :class:`int`

    object_id: :class:`int`

    object_type: :class:`str`
    """
    __slots__ = ("category", "id", "object_id", "object_type")

    def __init__(self, data):
        self.category = NotificationCategory(data["category"])
        self.id = data["id"]
        self.object_id = data["object_id"]
        self.object_type = ObjectType(data["object_type"])

    def __repr__(self):
        return prettify(self, 'id')


class NotificationsDetailsBase:
    """
    Base class for all Detail objects.

    **Attributes**

    username: :class:`str`
    """
    __slots__ = ("username",)

    def __init__(self, data):
        self.username = data["username"]

    def __repr__(self):
        # all the subclasses have a title attribute
        return prettify(self, "title", "username")


class BeatmapOwnerChangeDetails(NotificationsDetailsBase):
    """
    When ownership of a beatmap is transferred (for guest difficulties).

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    beatmap_id: :class:`int`

    cover_url: :class:`str`

    title: :class:`str`

    title_unicode: :class:`str`

    version: :class:`str`
    """
    __slots__ = ("beatmap_id", "cover_url", "title", "title_unicode", "version")

    def __init__(self, data):
        super().__init__(data)
        self.beatmap_id = data["beatmap_id"]
        self.cover_url = data["cover_url"]
        self.title = data["title"]
        self.title_unicode = data["title_unicode"]
        self.version = data["version"]


class BeatmapsetNotificationDetails(NotificationsDetailsBase):
    """
    Base class for several other detail objects.

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    title: :class:`str`

    title_unicode: :class:`str`

    cover_url: :class:`str`
    """
    __slots__ = ("title", "title_unicode", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.title = data["title"]
        self.title_unicode = data["title_unicode"]
        self.cover_url = data["cover_url"]


class BeatmapsetDiscussionPostNotificationDetails(NotificationsDetailsBase):
    """
    Base class for a couple other detail objects.

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    content: :class:`str`

    title: :class:`str`

    title_unicode: :class:`str`

    post_id: :class:`int`

    discussion_id: :class:`int`

    beatmap_id: :class:`int`

    cover_url: :class:`str`
    """
    __slots__ = (
        "content", "title", "title_unicode", "post_id", "discussion_id",
        "beatmap_id", "cover_url"
    )

    def __init__(self, data):
        super().__init__(data)
        self.content = data["content"]
        self.title = data["title"]
        self.title_unicode = data["title_unicode"]
        self.post_id = data["post_id"]
        self.discussion_id = data["discussion_id"]
        self.beatmap_id = data["beatmap_id"]
        self.cover_url = data["cover_url"]


class BeatmapsetDiscussionReviewNewDetails(NotificationsDetailsBase):
    """
    New beatmapset discussion review.

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    title: :class:`str`

    title_unicode: :class:`str`

    post_id: :class:`int`

    discussion_id: :class:`int`

    beatmap_id: :class:`int`

    cover_url: :class:`str`

    embeds: :class:`dict`
        contains keys 'suggestions', 'problems', and 'praises'
    """
    __slots__ = (
        "title", "title_unicode", "post_id", "discussion_id",
        "beatmap_id", "cover_url", "embeds"
    )

    def __init__(self, data):
        super().__init__(data)
        self.title = data["title"]
        self.title_unicode = data["title_unicode"]
        self.post_id = data["post_id"]
        self.discussion_id = data["discussion_id"]
        self.beatmap_id = data["beatmap_id"]
        self.cover_url = data["cover_url"]
        self.embeds = data["embeds"]


class ChannelAnnouncementDetails(NotificationsDetailsBase):
    """
    Chat channel announcement

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    channel_id: :class:`int`

    name: :class:`str`

    title: :class:`str`

    type: :class:`ChatChannelType`

    cover_url: :class:`str`
    """
    __slots__ = ("channel_id", "name", "title", "type", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.channel_id = data["channel_id"]
        self.name = data["name"]
        self.title = data["title"]
        self.type = ChatChannelType(data["type"].upper())
        self.cover_url = data["cover_url"]


class ChannelMessageDetails(NotificationsDetailsBase):
    """
    Chat channel message

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    title: :class:`str`

    type: :class:`ChatChannelType`

    cover_url: :class:`str`
    """
    __slots__ = ("title", "type", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.title = data["title"]
        self.type = ChatChannelType(data["type"].upper())
        self.cover_url = data["cover_url"]


class CommentNewDetails(NotificationsDetailsBase):
    """
    New comment

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    comment_id: :class:`int`

    title: :class:`str`

    content: :class:`str`

    cover_url: :class:`str`
    """
    __slots__ = ("comment_id", "title", "content", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.comment_id = data["comment_id"]
        self.title = data["title"]
        self.content = data["content"]
        self.cover_url = data["cover_url"]


class ForumTopicReplyDetails(NotificationsDetailsBase):
    """
    Forum topic reply

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    title: :class:`str`

    post_id: :class:`int`

    cover_url: :class:`str`
    """
    __slots__ = ("title", "post_id", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.title = data["title"]
        self.post_id = data["post_id"]
        self.cover_url = data["cover_url"]


class UserAchievementUnlockDetails(NotificationsDetailsBase):
    """
    New achievement unlocked

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    achievement_id: :class:`int`

    achievement_mode: Optional[:class:`GameModeStr`]

    cover_url: :class:`str`

    slug: :class:`str`

    title: :class:`str`

    user_id: :class:`int`
    """
    __slots__ = (
        "achievement_id", "achievement_mode", "cover_url",
        "slug", "title", "user_id"
    )

    def __init__(self, data):
        super().__init__(data)
        self.achievement_id = data["achievement_id"]
        self.achievement_mode = GameModeStr(data["achievement_mode"]) \
            if data["achievement_mode"] is not None else None
        self.cover_url = data["cover_url"]
        self.slug = data["slug"]
        self.title = data["title"]
        self.user_id = data["user_id"]


class UserBeatmapsetNewDetails(NotificationsDetailsBase):
    """
    New beatmapset

    Extends: :class:`NotificationsDetailsBase`

    **Attributes**

    beatmapset_id: :class:`int`

    title: :class:`str`

    title_unicode: :class:`str`

    cover_url: :class:`str`
    """
    __slots__ = ("beatmapset_id", "title", "title_unicode", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.beatmapset_id = data["beatmapset_id"]
        self.title = data["title"]
        self.title_unicode = data["title_unicode"]
        self.cover_url = data["cover_url"]
        self.username = data.get("username")


class BeatmapsetDiscussionLockDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset discussion locked

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetDiscussionPostNewDetails(BeatmapsetDiscussionPostNotificationDetails):
    """
    New beatmapset discussion post

    Extends :class:`BeatmapsetDiscussionPostNotificationDetails`
    """


class BeatmapsetDiscussionQualifiedProblemDetails(BeatmapsetDiscussionPostNotificationDetails):
    """
    Beatmapset discussion qualified problem

    Extends :class:`BeatmapsetDiscussionPostNotification`
    """


class BeatmapsetDiscussionUnlockDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset discussion unlocked

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetDisqualifyDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset disqualified

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetLoveDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset loved

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetNominateDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset nominated

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetQualifyDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset qualified

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetRankDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset ranked

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetRemoveFromLovedDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset removed from love

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetResetNominationsDetails(BeatmapsetNominateDetails):
    """
    Beatmapset nominations reset

    Extends :class:`BeatmapsetNominateDetails`
    """


class UserBeatmapsetReviveDetails(UserBeatmapsetNewDetails):
    """
    User beatmapset revived

    Extends :class:`UserBeatmapsetNewDetails`
    """


class Details:
    DETAIL_MAP = {
        NotificationType.BEATMAP_OWNER_CHANGE: BeatmapOwnerChangeDetails,
        NotificationType.BEATMAPSET_DISCUSSION_LOCK: BeatmapsetDiscussionLockDetails,
        NotificationType.BEATMAPSET_DISCUSSION_POST_NEW: BeatmapsetDiscussionPostNewDetails,
        NotificationType.BEATMAPSET_DISCUSSION_QUALIFIED_PROBLEM: BeatmapsetDiscussionQualifiedProblemDetails,
        NotificationType.BEATMAPSET_DISCUSSION_REVIEW_NEW: BeatmapsetDiscussionReviewNewDetails,
        NotificationType.BEATMAPSET_DISCUSSION_UNLOCK: BeatmapsetDiscussionUnlockDetails,
        NotificationType.BEATMAPSET_DISQUALIFY: BeatmapsetDisqualifyDetails,
        NotificationType.BEATMAPSET_LOVE: BeatmapsetLoveDetails,
        NotificationType.BEATMAPSET_NOMINATE: BeatmapsetNominateDetails,
        NotificationType.BEATMAPSET_QUALIFY: BeatmapsetQualifyDetails,
        NotificationType.BEATMAPSET_RANK: BeatmapsetRankDetails,
        NotificationType.BEATMAPSET_REMOVE_FROM_LOVED: BeatmapsetRemoveFromLovedDetails,
        NotificationType.BEATMAPSET_RESET_NOMINATIONS: BeatmapsetResetNominationsDetails,
        NotificationType.CHANNEL_ANNOUNCEMENT: ChannelAnnouncementDetails,
        NotificationType.CHANNEL_MESSAGE: ChannelMessageDetails,
        NotificationType.COMMENT_NEW: CommentNewDetails,
        NotificationType.FORUM_TOPIC_REPLY: ForumTopicReplyDetails,
        NotificationType.USER_ACHIEVEMENT_UNLOCK: UserAchievementUnlockDetails,
        NotificationType.USER_BEATMAPSET_NEW: UserBeatmapsetNewDetails,
        NotificationType.USER_BEATMAPSET_REVIVE: UserBeatmapsetReviveDetails,
    }

    def __new__(cls, data, detail_type):
        return cls.DETAIL_MAP[detail_type](data)



from dateutil import parser

from ..util import prettify
from ..enums import GameModeStr


class Event:
    """
    The object has different attributes depending on its type. Following are attributes available to all types.

    **Attributes**

    created_at: :class:`datetime.datetime`

    id: :class:`int`

    type:
        All types and the additional attributes they provide are listed under 'Event Types'

    **Event Types**

    :class:`achievement`
        achievement: :class:`Achievement`
        user: :class:`EventUser`

    :class:`beatmapPlaycount`
        beatmap: :class:`EventBeatmap`
        count: :class:`int`

    :class:`beatmapsetApprove`
        approval: :class:`str`
        beatmapset: :class:`EventBeatmapset`
        user: :class:`EventUser`

    :class:`beatmapsetDelete`
        beatmapset: :class:`EventBeatmapset`

    :class:`beatmapsetRevive`
        beatmapset: :class:`EventBeatmapset`
        user: :class:`EventUser`

    :class:`beatmapsetUpdate`
        beatmapset: :class:`EventBeatmapset`
        user: :class:`EventUser`

    :class:`beatmapsetUpload`
        beatmapset: :class:`EventBeatmapset`
        user: :class:`EventUser`

    :class:`rank`
        score_rank: :class:`str`
        rank: :class:`int`
        mode: :class:`GameModeStr`
        beatmap: :class:`EventBeatmap`
        user: :class:`EventUser`

    :class:`rankLost`
        mode: :class:`GameModeStr`
        beatmap: :class:`EventBeatmap`
        user: :class:`EventUser`

    :class:`userSupportAgain`
        user: :class:`EventUser`

    :class:`userSupportFirst`
        user: :class:`EventUser`

    :class:`userSupportGift`
        user: :class:`EventUser`

    :class:`usernameChange`
        user: :class:`EventUser`
    """
    def __init__(self, data):
        self.created_at = parser.parse(data['created_at'])
        self.id = data['id']
        self.type = data['type']

        if self.type == 'achievement':
            self.achievement = data['achievement']
            self.user = EventUser(data['user'])
        elif self.type == 'beatmapPlaycount':
            self.beatmap = EventBeatmap(data['beatmap'])
            self.count = data['count']
        elif self.type == 'beatmapsetApprove':
            self.approval = data['approval']
            self.beatmapset = EventBeatmapset(data['beatmapset'])
            self.user = EventUser(data['user'])
        elif self.type == 'beatmapsetDelete':
            self.beatmapset = EventBeatmapset(data['beatmapset'])
        elif self.type in ("beatmapsetRevive", "beatmapsetUpdate", "beatmapsetUpload"):
            self.beatmapset = EventBeatmapset(data['beatmapset'])
            self.user = EventUser(data['user'])
        elif self.type == 'rank':
            self.score_rank = data['scoreRank']
            self.rank = data['rank']
            self.mode = GameModeStr(data['mode'])
            self.beatmap = EventBeatmap(data['beatmap'])
            self.user = EventUser(data['user'])
        elif self.type == 'rankLost':
            self.mode = GameModeStr(data['mode'])
            self.beatmap = EventBeatmap(data['beatmap'])
            self.user = EventUser(data['user'])
        elif self.type in ("userSupportAgain", "userSupportFirst", "userSupportGift", "usernameChange"):
            self.user = EventUser(data['user'])

    def __repr__(self):
        slots = ['type', 'user', 'beatmapset']
        fields = [slot for slot in slots if getattr(self, slot, None)]
        return prettify(self, *fields)


class EventUser:
    """
    **Attributes**

    username: :class:`str`

    url: :class:`str`

    previous_username: :class:`str` or :class:`NoneType`
    """
    __slots__ = (
        "username", "url", "previous_username"
    )

    def __init__(self, data):
        self.username = data['username']
        self.url = data['url']
        self.previous_username = data.get('previousUsername', None)

    def __repr__(self):
        return prettify(self, 'username', 'url')


class EventBeatmap:
    """
    **Attributes**

    title: :class:`str`

    url: :class:`str`
    """
    __slots__ = (
        "title", "url"
    )

    def __init__(self, data):
        self.title = data['title']
        self.url = data['url']

    def __repr__(self):
        return prettify(self, 'title', 'url')


class EventBeatmapset:
    """
    **Attributes**

    title: :class:`str`

    url: :class:`str`
    """
    __slots__ = (
        "title", "url"
    )

    def __init__(self, data):
        self.title = data['title']
        self.url = data['url']

    def __repr__(self):
        return prettify(self, 'title', 'url')



from .user import UserCompact
from dateutil import parser

from ..util import prettify


class Comment:
    """
    Represents a single comment.

    **Attributes**

    commentable_id: :class:`int`
        ID of the object the comment is attached to

    commentable_type: :class:`str`
        type of object the comment is attached to

    created_at: :class:`datetime.datetime`
        ISO 8601 date

    deleted_at: Union[:class:`datetime.datetime`, :class:`None`]
        ISO 8601 date if the comment was deleted; null, otherwise

    edited_at: Union[:class:`datetime.datetime`, :class:`None`]
        ISO 8601 date if the comment was edited; null, otherwise

    edited_by_id: Union[:class:`int`, :class:`None`]
        user id of the user that edited the post; null, otherwise

    id: :class:`int`
        the ID of the comment

    legacy_name: Union[:class:`str`, :class:`None`]
        username displayed on legacy comments

    message: Union[:class:`str`, :class:`None`]
        markdown of the comment's content

    message_html: Union[:class:`str`, :class:`None`]
        html version of the comment's content

    parent_id: Union[:class:`int`, :class:`None`]
        ID of the comment's parent

    pinned: :class:`bool`
        Pin status of the comment

    replies_count: :class:`int`
        number of replies to the comment

    updated_at: :class:`datetime.datetime`
        ISO 8601 date

    user_id: :class:`int`
        user ID of the poster

    votes_count: :class:`int`
        number of votes

    url: :class:`str`
        URL to the comment
    """
    __slots__ = (
        "commentable_id", "commentable_type", "created_at", "deleted_at", "edited_at", "edited_by_id",
        "id", "legacy_name", "message", "message_html", "parent_id", "pinned", "replies_count", "updated_at",
        "user_id", "votes_count"
    )

    def __init__(self, data):
        self.commentable_id = data['commentable_id']
        self.commentable_type = data['commentable_type']
        self.created_at = parser.parse(data['created_at'])
        self.deleted_at = parser.parse(data['deleted_at']) if data['deleted_at'] is not None else None
        self.edited_at = parser.parse(data['edited_at']) if data['edited_at'] is not None else None
        self.edited_by_id = data['edited_by_id']
        self.id = data['id']
        self.legacy_name = data['legacy_name']
        self.message = data.get('message')
        self.message_html = data.get('message_html')
        self.parent_id = data['parent_id']
        self.pinned = data['pinned']
        self.replies_count = data['replies_count']
        self.updated_at = data['updated_at']
        self.user_id = data['user_id']
        self.votes_count = data['votes_count']

    @property
    def url(self):
        return f"https://osu.ppy.sh/comments/{self.id}"

    def __repr__(self):
        return prettify(self, 'user_id', 'message')


class CommentBundle:
    """
    Comments and related data.

    **Attributes**

    commentable_meta: Sequence[:class:`CommentableMeta`]
        ID of the object the comment is attached to

    comments: Sequence[:class:`Comment`]
        List of comments ordered according to sort

    cursor:	:class:`Cursor`
        To be used to query the next page

    has_more: :class:`bool`
        If there are more comments or replies available

    has_more_id: Union[:class:`id`, :class:`None`]

    included_comments: Sequence[:class:`Comment`]
        Related comments; e.g. parent comments and nested replies

    pinned_comments: Union[Sequence[:class:`Comment`], :class:`None`]
        Pinned comments

    sort: :class:`str`
        one of the following:
            new (created_at (descending), id (descending))
            old (created_at (ascending), id (ascending))
            top (votes_count (descending), created_at (descending), id (descending))

    top_level_count: Union[:class:`int`, :class:`None`]
        Number of comments at the top level. Not returned for replies.

    total: Union[:class:`int`, :class:`None`]
        Total number of comments. Not retuned for replies.

    user_follow: :class:`bool`
        is the current user watching the comment thread?

    user_votes: Sequence[:class:`int`]
        IDs of the comments in the bundle the current user has upvoted

    users: Sequence[:class:`UserCompact`]
        list containing objects of type :class:`UserCompact`. list of users related to the comments
    """
    __slots__ = (
        "commentable_meta", "comments", "cursor", "has_more", "has_more_id", "included_comments",
        "pinned_comments", "sort", "top_level_count", "total", "user_follow", "user_votes", "users"
    )

    def __init__(self, data):
        self.commentable_meta = list(map(CommentableMeta, data['commentable_meta']))
        self.comments = list(map(Comment, data['comments']))
        self.cursor = data['cursor']
        self.has_more = data['has_more']
        self.has_more_id = data.get('has_more_id')
        self.included_comments = list(map(Comment, data['included_comments']))
        self.pinned_comments = list(map(Comment, data['pinned_comments'])) \
            if data.get('pinned_comments') is not None else None
        self.sort = data['sort']
        self.top_level_count = data.get('top_level_count')
        self.total = data.get('total')
        self.user_follow = data['user_follow']
        self.user_votes = data['user_votes']
        self.users = list(map(UserCompact, data['users']))

    def __repr__(self):
        return prettify(self, 'commentable_meta', 'comments', 'users')


class CommentableMeta:
    """
    Metadata of the object that a comment is attached to.
    Not all attributes are guaranteed to have a value, for example
    if the comment is delete.

    **Attributes**

    id: :class:`int`
        the ID of the object

    title: :class:`str`
        display title

    type: :class:`str`
        the type of the object

    url: :class:`str`
        url of the object

    owner_id: :class:`int`:
        the ID of the owner of the object

    owner_title: :class:`int`
        undocumented
    """
    __slots__ = (
        "id", "title", "type", "url", "owner_id", "owner_title"
    )

    def __init__(self, data):
        self.id = data.get('id')
        self.title = data.get('title')
        self.type = data.get('type')
        self.url = data.get('url')

    def __repr__(self):
        return prettify(self, 'title', 'url')



import requests
import time

from .exceptions import ScopeException
from .constants import base_url, lazer_base_url


class HTTPHandler:
    def __init__(self, client, request_wait_time, limit_per_minute, use_lazer=False):
        self.client = client
        self.rate_limit = RateLimitHandler(request_wait_time, limit_per_minute)
        self.use_lazer = use_lazer

    def get_headers(self, requires_auth=True, is_files=False, **kwargs):
        headers = {
            'charset': 'utf-8',
            **{str(key): str(value) for key, value in kwargs.items() if value is not None}
        }
        if not is_files and "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"
        if requires_auth and "Authorization" not in headers:
            headers['Authorization'] = f"Bearer {self.client.auth.token}"
        return headers

    def make_request(self, path, data=None, headers=None, is_download=False, files=None, **kwargs):
        if headers is None:
            headers = {}
        if data is None:
            data = {}

        if path.requires_auth and self.client.auth is None:
            raise ScopeException("You need to be authenticated to make this request.")

        if path.requires_auth and path.scope not in self.client.auth.scope:
            raise ScopeException(f"You don't have the {path.scope} scope, which is required to make this request.")

        if path.requires_user and not self.client.auth.has_user:
            raise ScopeException("This request requires a user. You need either a delegate scope or "
                                 "to register OAuth with Authorization Code Grant.")

        if not self.rate_limit.can_request:
            self.rate_limit.wait()

        headers = self.get_headers(path.requires_auth, files is not None, **headers)
        params = {str(key): value for key, value in kwargs.items() if value is not None}
        response = getattr(requests, path.method)((lazer_base_url if self.use_lazer else base_url) + path.path,
                                                  headers=headers, data=data, params=params, files=files)
        self.rate_limit.request_used()
        try:
            response.raise_for_status()
        except Exception as e:
            try:
                err = response.json()['error']
            except:
                err = None
            raise type(e)(str(e)+": "+err) if err is not None else e
        if response.content == b"":
            return
        return response.json() if not is_download else response.content


class RateLimitHandler:
    def __init__(self, request_wait_limit, limit_per_minute):
        self.wait_limit = request_wait_limit
        self.limit = limit_per_minute
        self.last_request = time.perf_counter() - self.wait_limit
        self.requests = []

    def request_used(self):
        self.requests.append(time.perf_counter())
        self.last_request = time.perf_counter()

    def wait(self):
        next_available_request = self.wait_limit-(time.perf_counter()-self.last_request)
        if len(self.requests) >= self.limit:
            next_available_request = max(next_available_request, self.requests[0]+60-time.perf_counter())
        time.sleep(next_available_request)

    def reset(self):
        while len(self.requests) > 0:
            if self.requests[0] + 60 < time.perf_counter():
                self.requests.pop(0)
            else:
                break

    @property
    def can_request(self):
        self.reset()
        return time.perf_counter()-self.last_request >= self.wait_limit and len(self.requests) < self.limit



from .http import HTTPHandler
from .objects import *
from .path import Path
from .enums import *
from .auth import AuthHandler, LazerAuthHandler
from .util import (
    parse_mods_arg, parse_enum_args,
    BeatmapsetSearchFilter, create_multipart_formdata,
    PlaylistItemUtil, IdentitiesUtil, NotificationsUtil,
    JsonUtil
)
from typing import Union, Optional, Sequence, Dict
from datetime import datetime
from osrparse import Replay
import json


class Client:
    """
    Main object for interacting with osu!api

    **Init Parameters**

    auth: :class:`AuthHandler`
        The AuthHandler object passed in when initiating the Client object

    request_wait_time: Optional[:class:`float`]
        Default is 1.

        This defines the amount of time that the client should wait before making another request.
        It can make it easier to stay within the rate limits without using all your requests up quickly
        and then waiting forever to make another. It's most applicable in bot-type apps.

    limit_per_minute: Optional[:class:`float`]
        Default is 60 because that's the limit peppy requests that we stay under.

        This sets a cap on the number of requests the client is allowed to make within 1 minute of time.

    use_lazer: Optional[:class:`bool`]
        Default is False. This changes which base api endpoint the client will use.

        Uses lazer.ppy.sh when True and osu.ppy.sh when False.

    Make sure if you are changing the ratelimit handling that you are still following peppy's
    TOU for using the API:

    Use the API for good. Don't overdo it. If in doubt, ask before (ab)using :).
    this section may expand as necessary.

    Current rate limit is set at an insanely high 1200 requests per minute,
    with burst capability of up to 200 beyond that.
    If you require more, you probably fall into the above category of abuse.
    If you are doing more than 60 requests a minute,
    you should probably give peppy a yell.
    """
    def __init__(self, auth: Union[AuthHandler, LazerAuthHandler] = None, request_wait_time: Optional[float] = 1.0,
                 limit_per_minute: Optional[float] = 60.0, use_lazer: Optional[bool] = False):
        self.auth = auth
        self.http = HTTPHandler(self, request_wait_time, limit_per_minute, use_lazer)

    @classmethod
    def from_client_credentials(cls, client_id: int, client_secret: str, redirect_url: str,
                                scope: Optional[Scope] = Scope.default(), code: Optional[str] = None,
                                request_wait_time: Optional[float] = 1.0,
                                limit_per_minute: Optional[float] = 60.0):
        """
        Returns a :class:`Client` object from client id, client secret, redirect uri, and scope.

        **Parameters**

        client_id: :class:`int`
            API Client id

        client_secret: :class:`int`
            API Client secret

        redirect_uri: :class:`str`
            API redirect uri

        scope: Optional[:class:`Scope`]
            Scopes to use. Default is Scope.default() which is just the public scope.

        code: Optional[:class:`str`]
            If provided, is used to authorize. Read more about this under :class:`AuthHandler.get_auth_token`

        request_wait_time: Optional[:class:`float`]
            Read under Client init parameters.

        limit_per_minute: Optional[:class:`float`]
            Read under Client init parameters.

        **Returns**

        :class:`Client`
        """
        auth = AuthHandler(client_id, client_secret, redirect_url, scope)
        auth.get_auth_token(code)
        return cls(auth, request_wait_time, limit_per_minute)

    @classmethod
    def from_osu_credentials(cls, username: str, password: str,
                             request_wait_time: Optional[float] = 1.0,
                             limit_per_minute: Optional[float] = 60.0):
        """
        Returns a :class:`Client` object which will make authorize and make requests to
        lazer.ppy.sh

        username: :class:`str`
            osu! username login

        password: :class:`str`
            osu! password login

        request_wait_time: Optional[:class:`float`]
            Read under Client init parameters.

        limit_per_minute: Optional[:class:`float`]
            Read under Client init parameters.
        """
        auth = LazerAuthHandler(username, password)
        auth.get_auth_token()
        return cls(auth, request_wait_time, limit_per_minute, True)

    def lookup_beatmap(self, checksum: Optional[str] = None, filename: Optional[str] = None,
                       id: Optional[int] = None) -> Beatmap:
        """
        Returns beatmap.

        Requires OAuth and scope public

        **Parameters**

        checksum: Optional[:class:`str`]
            A beatmap checksum.

        filename: Optional[:class:`str`]
            A filename to lookup

        id: Optional[:class:`int`]
            A beatmap ID to lookup

        **Returns**

        :class:`Beatmap`
        """
        return Beatmap(self.http.make_request(Path.beatmap_lookup(), checksum=checksum,
                                              filename=filename, id=id))

    def get_user_beatmap_score(self, beatmap: int, user: int, mode: Optional[Union[str, GameModeStr]] = None,
                               mods: Optional[Sequence[str]] = None) -> BeatmapUserScore:
        """
        Returns a user's score on a Beatmap

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Id of the beatmap

        user: :class:`int`
            Id of the user

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            The game mode to get scores for

        mods: Optional[Sequence[:class:`str`]]
            An array of matching mods, or none. Currently doesn't do anything.

        **Returns**

        :class:`BeatmapUserScore`
        """
        mode = parse_enum_args(mode)
        return BeatmapUserScore(self.http.make_request(Path.user_beatmap_score(beatmap, user),
                                                       mode=mode, mods=mods))

    def get_user_beatmap_scores(self, beatmap: int, user: int,
                                mode: Optional[Union[str, GameModeStr]] = None) -> Sequence[LegacyScore]:
        """
        Returns a user's scores on a Beatmap

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Id of the beatmap

        user: :class:`int`
            Id of the user

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            The game mode to get scores for

        **Returns**

        Sequence[:class:`LegacyScore`]
        """
        mode = parse_enum_args(mode)
        return list(map(LegacyScore, self.http.make_request(Path.user_beatmap_scores(beatmap, user),
                                                            mode=mode)["scores"]))

    def get_beatmap_scores(self, beatmap: int, mode: Optional[Union[str, GameModeStr]] = None,
                           mods: Optional[Sequence[str]] = None,
                           type: Optional[Sequence[str]] = None) -> BeatmapScores:
        """
        Returns the top scores for a beatmap

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Id of the beatmap

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            The game mode to get scores for

        mods: Optional[Sequence[:class:`str`]]
            An array of matching mods, or none. Currently doesn't do anything.

        type: Optional[Sequence[:class:`str`]]
            Beatmap score ranking type. Currently doesn't do anything.

        **Returns**

        :class:`BeatmapScores`
            :class:`LegacyScore` object inside includes "user" and the included user includes "country" and "cover".
        """
        mode = parse_enum_args(mode)
        return BeatmapScores(self.http.make_request(Path.beatmap_scores(beatmap), mode=mode,
                                                    mods=mods, type=type))

    def get_lazer_beatmap_scores(self, beatmap: int, mode: Optional[Union[str, GameModeStr]] = None,
                                 mods: Optional[Sequence[str]] = None,
                                 type: Optional[Sequence[str]] = None) -> BeatmapScores:
        """
        Returns the top scores for a beatmap on the lazer client.

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Id of the beatmap

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            The game mode to get scores for

        mods: Optional[Sequence[:class:`str`]]
            An array of matching mods, or none. Currently doesn't do anything.

        type: Optional[Sequence[:class:`str`]]
            Beatmap score ranking type. Currently doesn't do anything.

        **Returns**

        :class:`BeatmapScores`
            :class:`LazerScore` object inside includes "user" and the included user includes "country" and "cover".
        """
        mode = parse_enum_args(mode)
        return BeatmapScores(self.http.make_request(Path.lazer_beatmap_scores(beatmap), mode=mode,
                                                    mods=mods, type=type), "lazer")

    def get_beatmap(self, beatmap: int) -> Beatmap:
        """
        Gets beatmap data for the specified beatmap ID.

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            The ID of the beatmap

        **Returns**

        :class:`Beatmap`
            Includes attributes beatmapset, failtimes, and max_combo
        """
        return Beatmap(self.http.make_request(Path.beatmap(beatmap)))

    def get_beatmaps(self, ids: Optional[Sequence[int]] = None) -> Sequence[Beatmap]:
        """
        Returns list of beatmaps.

        Requires OAuth and scope public

        **Parameters**

        ids: Optional[List[:class:`int`]]
            Beatmap id to be returned. Specify once for each beatmap id requested.
            Up to 50 beatmaps can be requested at once.

        **Returns**

        Sequence[:class:`BeatmapCompact`]
            Includes: beatmapset (with ratings), failtimes, max_combo.
        """
        results = self.http.make_request(Path.beatmaps(), **{"ids[]": ids})
        return list(map(Beatmap, results['beatmaps'])) if results else []

    def get_beatmap_attributes(self, beatmap: int,
                               mods: Optional[Union[int, Mods, Sequence[Union[str, Mods, int]]]] = None,
                               ruleset: Optional[Union[str, GameModeStr]] = None,
                               ruleset_id: Optional[Union[int, GameModeInt]] = None) -> BeatmapDifficultyAttributes:
        """
        Returns difficulty attributes of beatmap with specific mode and mods combination.

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Beatmap id.

        mods: Optional[Union[:class:`int`, Sequence[Union[:class:`str`, :class:`Mods`, :class:`int`]], :class:`Mods`]]
            Mod combination. Can be either a bitset of mods, a Mods enum, or array of any. Defaults to no mods.
            Some mods may cause the api to throw an HTTP 422 error depending on the map's gamemode.

        ruleset: Optional[Union[:class:`GameModeStr`, :class:`int`]]
            Ruleset of the difficulty attributes. Only valid if it's the beatmap ruleset or the beatmap can be
            converted to the specified ruleset. Defaults to ruleset of the specified beatmap.

        ruleset_id: Optional[Union[:class:`GameModeInt`, :class:`int`]]
            The same as ruleset but in integer form.

        **Returns**

        :class:`BeatmapDifficultyAttributes`
        """
        ruleset, ruleset_id = parse_enum_args(ruleset, ruleset_id)
        return BeatmapDifficultyAttributes(self.http.make_request(Path.get_beatmap_attributes(beatmap),
                                                                  mods=parse_mods_arg(mods), ruleset=ruleset,
                                                                  ruleset_id=ruleset_id))

    def get_beatmapset(self, beatmapset_id: int) -> Beatmapset:
        """
        Get beatmapset by id.

        Requires OAuth and scope public

        **Parameters**

        beatmapset_id: :class:`int`

        **Returns**

        :class:`Beatmapset`
        """
        return Beatmapset(self.http.make_request(Path.get_beatmapset(beatmapset_id)))

    def get_beatmapset_discussion_posts(self, beatmapset_discussion_id: Optional[int] = None,
                                        limit: Optional[int] = None, page: Optional[int] = None,
                                        sort: Optional[str] = None, user: Optional[int] = None,
                                        with_deleted: Optional[str] = None) -> dict:
        """
        Returns the posts of the beatmapset discussions

        Requires OAuth and scope public

        **Parameters**

        beatmapset_discussion_id: Optional[:class:`int`]
            id of the BeatmapsetDiscussion

        limit: Optional[:class:`int`]
            Maximum number of results

        page: Optional[:class:`int`]
            Search results page.

        sort: Optional[:class:`str`]
            id_desc for newest first; id_asc for oldest first. Defaults to id_desc

        user: Optional[:class:`int`]
            The id of the User

        with_deleted: Optional[:class:`str`]
            The param has no effect as api calls do not currently receive group permissions

        **Returns**

        :class:`dict`
            {
            beatmapsets: :class:`BeatmapsetCompact`,

            cursor: :class:`dict`,

            posts: Sequence[:class:`BeatmapsetDiscussionPost`],

            users: :class:`UserCompact`
            }
        """
        # TODO: Change is supposed to occur on the response given back from the server,
        #  make sure to change it when that happens.
        resp = self.http.make_request(Path.beatmapset_discussion_posts(),
                                      beatmapset_discussion_id=beatmapset_discussion_id,
                                      limit=limit, page=page, sort=sort, user=user, with_deleted=with_deleted)
        return {
            'beatmapsets': list(map(BeatmapsetCompact, resp['beatmapsets'])),
            'cursor': resp['cursor'],
            'posts': list(map(BeatmapsetDiscussionPost, resp['posts'])),
            'users': list(map(UserCompact, resp['users']))
        }

    def get_beatmapset_discussion_votes(self, beatmapset_discussion_id: Optional[int] = None,
                                        limit: Optional[int] = None, page: Optional[int] = None,
                                        receiver: Optional[int] = None, score: Optional[int] = None,
                                        sort: Optional[str] = None, user: Optional[int] = None,
                                        with_deleted: Optional[str] = None) -> dict:
        """
        Returns the votes given to beatmapset discussions

        Requires OAuth and scope public

        **Parameters**

        beatmapset_discussion_id: Optional[:class:`int`]
            id of the BeatmapsetDiscussion

        limit: Optional[:class:`int`]
            Maximum number of results

        page: Optional[:class:`int`]
            Search results page.

        receiver: Optional[:class:`int`]
            The id of the User receiving the votes.

        score: Optional[:class:`int`]
            1 for upvote, -1 for downvote

        sort: Optional[:class:`str`]
            id_desc for newest first; id_asc for oldest first. Defaults to id_desc

        user: Optional[:class:`int`]
            The id of the User giving the votes.

        with_deleted: Optional[:class:`str`]
            The param has no effect as api calls do not currently receive group permissions

        **Returns**

        :class:`dict`
            {
            cursor: :class:`dict`,

            discussions: Sequence[:class:`BeatmapsetDiscussion`],

            users: Sequence[:class:`UserCompact`],

            votes: Sequence[:class:`BeatmapsetDiscussionVote`]
            }
        """
        # TODO: Change is supposed to occur on the response given back from the server,
        #  make sure to change it when that happens.
        resp = self.http.make_request(Path.beatmapset_discussion_votes(),
                                      beatmapset_discussion_id=beatmapset_discussion_id,
                                      limit=limit, receiver=receiver, score=score, page=page,
                                      sort=sort, user=user, with_deleted=with_deleted)
        return {
            'cursor': resp['cursor'],
            'discussions': list(map(BeatmapsetDiscussion, resp['discussions'])),
            'users': list(map(UserCompact, resp['users'])),
            'votes': list(map(BeatmapsetDiscussionVote, resp['votes']))
        }

    def get_beatmapset_discussions(self, beatmap_id: Optional[int] = None, beatmapset_id: Optional[int] = None,
                                   beatmapset_status: Optional[str] = None, limit: Optional[int] = None,
                                   message_types: Optional[Sequence[str]] = None,
                                   only_unresolved: Optional[bool] = None, page: Optional[int] = None,
                                   sort: Optional[str] = None, user: Optional[int] = None,
                                   with_deleted: Optional[str] = None) -> dict:
        """
        Returns a list of beatmapset discussions

        Requires OAuth and scope public

        **Parameters**

        beatmap_id: Optional[:class:`int`]
            id of the Beatmap

        beatmapset_id: Optional[:class:`int`]
            id of the Beatmapset

        beatmapset_status: Optional[:class:`str`]
            One of all, ranked, qualified, disqualified, never_qualified. Defaults to all.

        limit: Optional[:class:`int`]
            Maximum number of results.

        message_types: Optional[Sequence[:class:`str`]]
            suggestion, problem, mapper_note, praise, hype, review. Blank defaults to all types.

        only_unresolved: Optional[:class:`bool`]
            true to show only unresolved issues; false, otherwise. Defaults to false.

        page: Optional[:class:`int`]
            Search result page.

        sort: Optional[:class:`str`]
            id_desc for newest first; id_asc for oldest first. Defaults to id_desc.

        user: Optional[:class:`int`]
            The id of the User.

        with_deleted: Optional[:class:`str`]
            This param has no effect as api calls do not currently receive group permissions.

        **Returns**

        :class:`dict`
            {

            beatmaps: Sequence[:class:`Beatmap`],
                List of beatmaps associated with the discussions returned.

            cursor: :class:`dict`,

            discussions: Sequence[:class:`BeatmapsetDiscussion`],
                List of discussions according to sort order.

            included_discussions: Sequence[:class:`BeatmapsetDiscussion`],
                Additional discussions related to discussions.

            reviews_config.max_blocks: :class:`int`,
                Maximum number of blocks allowed in a review.

            users: Sequence[:class:`UserCompact`]
                List of users associated with the discussions returned.

            }
        """
        # TODO: Change is supposed to occur on the response given back from the server,
        #  make sure to change it when that happens.
        message_types = {"message_types[]": message_types}
        resp = self.http.make_request(Path.beatmapset_discussions(), beatmap_id=beatmap_id,
                                      beatmapset_id=beatmapset_id, beatmapset_status=beatmapset_status,
                                      limit=limit, only_unresolved=only_unresolved, page=page, sort=sort,
                                      user=user, with_deleted=with_deleted, **message_types)
        return {
            'beatmaps': list(map(Beatmap, resp['beatmaps'])),
            'cursor': resp['cursor'],
            'discussions': list(map(BeatmapsetDiscussion, resp['discussions'])),
            'included_discussions': list(map(BeatmapsetDiscussion, resp['included_discussions'])),
            'reviews_config.max_blocks': resp['reviews_config'],
            'users': list(map(UserCompact, resp['users']))
        }

    def get_changelog_build(self, stream: str, build: str) -> Build:
        """
        Returns details of the specified build.

        **Parameters**

        stream: :class:`str`
            Update stream name.

        build: :class:`str`
            Build version.

        **Returns**

        A :class:`Build` with changelog_entries, changelog_entries.github_user, and versions included.
        """
        return Build(self.http.make_request(Path.get_changelog_build(stream, build)))

    def get_changelog_listing(self, from_version: Optional[str] = None, max_id: Optional[int] = None,
                              stream: Optional[str] = None, to: Optional[str] = None,
                              message_formats: Optional[Sequence[str]] = None) -> \
            Dict[str, Union[Sequence[Build], Sequence[UpdateStream], Dict[str, Union[str, int, None]]]]:
        """
        Returns a listing of update streams, builds, and changelog entries.

        **Parameters**

        from_version: Optional[:class:`str`]
            Minimum build version.

        max_id: Optional[:class:`int`]
            Maximum build ID.

        stream: Optional[:class:`str`]
            Stream name to return builds from.

        to: Optional[:class:`str`]
            Maximum build version.

        message_formats: Optional[Sequence[:class:`str`]]
            html, markdown. Default to both.

        **Returns**

        {

        "build": Sequence[:class:`Build`]

        "search": {

            "from": :class:`str`
                from_version input.

            "limit": :class:`int`
                Always 21.

            "max_id": :class:`int`
                max_id input.

            "stream": :class:`str`
                stream input.

            "to": :class:`str`
                to input.

        }

        "streams": Sequence[:class:`UpdateStream`]

        }
        """
        response = self.http.make_request(Path.get_changelog_listing(), max_id=max_id,
                                          stream=stream, to=to, message_formats=message_formats,
                                          **{"from": from_version})
        return {
            "build": list(map(Build, response['builds'])),
            "search": response['search'],
            "streams": list(map(UpdateStream, response['streams'])),
        }

    def lookup_changelog_build(self, changelog: str, key: Optional[str] = None,
                               message_formats: Optional[Sequence[str]] = None) -> Build:
        """
        Returns details of the specified build.

        **Parameters**

        changelog: :class:`str`
            Build version, update stream name, or build ID.

        key: Optional[:class:`str`]
            Unset to query by build version or stream name, or id to query by build ID.

        message_formats: Optional[Sequence[:class:`str`]]
            html, markdown. Default to both.

        **Returns**

        A :class:`Build` with changelog_entries, changelog_entries.github_user, and versions included.
        """
        return Build(self.http.make_request(Path.lookup_changelog_build(changelog),
                                            key=key, message_formats=message_formats))

    def chat_acknowledge(self, history_since: Optional[int] = None, since: Optional[int] = None) -> \
            Sequence[UserSilence]:
        """
        Send a chat ack.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        history_since: Optional[:class:`int]
            :class:`UserSilence`s after the specified id to return.
            This field is preferred and takes precedence over since.

        since: Optional[:class:`int`]
            :class:`UserSilence`s after the specified :class:`ChatMessage`.message_id to return.

        **Returns**

        Sequence[:class:`UserSilence`]
        """
        resp = self.http.make_request(Path.chat_ack(), history_since=history_since, since=since)
        return list(map(UserSilence, resp["silences"]))

    def create_new_pm(self, target_id: int, message: str, is_action: bool, uuid: Optional[str] = None) -> dict:
        """
        This endpoint allows you to create a new PM channel.

        Requires OAuth, scope chat.write, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        target_id: :class:`int`
            user_id of user to start PM with

        message: :class:`str`
            message to send

        is_action: :class:`bool`
            whether the message is an action

        uuid: Optional[:class:`str`]
            client-side message identifier which will be sent back in response and websocket json.

        **Returns**

        :class:`dict`
            {

            channel: :class:`ChatChannel`
                channel the message was sent in

            message: :class:`ChatMessage`
                the sent message

            new_channel_id: :class:`int`
                channel_id of newly created channel

            }
        """
        data = {'target_id': target_id, 'message': message, 'is_action': is_action}
        if uuid is not None:
            data['uuid'] = uuid
        resp = self.http.make_request(Path.create_new_pm(), files=create_multipart_formdata(data))
        return {
            'channel': ChatChannel(resp['channel']),
            'message': ChatMessage(resp['message']),
            'new_channel_id': resp['new_channel_id'],
        }

    def get_updates(self, since: int, includes: Optional[Sequence[str]] = None,
                    history_since: Optional[int] = None) -> dict:
        """
        This endpoint returns new messages since the given message_id along with updated channel 'presence' data.

        Requires OAuth, scope lazer, a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        since: :class:`int`
            Messages after the specified message_id to return.

        includes: Optional[Sequence[:class:`str`]]
            List of fields from presence, silences to include in the response. Returns presence if not specified.
            Specifying silences may cause the api to return blank content for both presence and silences.

        history_since: Optional[:class:`int`]
            :class:`UserSilence`s after the specified id to return.

        **Returns**

        :class:`dict`
            {
            presence: List[:class:`ChatChannel`],

            silences: List[:class:`UserSilence`]

            }
        """
        if includes is None:
            includes = ["presence"]
        resp = self.http.make_request(Path.get_updates(), since=since, history_since=history_since,
                                      **{"includes[]": ",".join(includes)})
        if resp is None:
            resp = {}
        return {
            'presence': list(map(ChatChannel, resp['presence'])) if 'presence' in resp else None,
            'silences': list(map(UserSilence, resp['silences'])) if 'silences' in resp else None
        }

    def get_channel_messages(self, channel_id: int, limit: Optional[int] = None, since: Optional[int] = None,
                             until: Optional[int] = None) -> Sequence[ChatMessage]:
        """
        This endpoint returns the chat messages for a specific channel.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameter**

        channel_id: :class:`int`
            The ID of the channel to retrieve messages for

        limit: Optional[:class:`int`]
            number of messages to return (max of 50)

        since: Optional[:class:`int`]
            messages after the specified message id will be returned

        until: Optional[:class:`int`]
            messages up to but not including the specified message id will be returned

        **Returns**

        Sequence[:class:`ChatMessage`]
            list containing :class:`ChatMessage` objects
        """
        return list(map(ChatMessage, self.http.make_request(Path.get_channel_messages(channel_id),
                                                            limit=limit, since=since, until=until)))

    def send_message_to_channel(self, channel_id: int, message: str, is_action: bool,
                                uuid: Optional[str] = None) -> ChatMessage:
        """
        This endpoint sends a message to the specified channel.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        channel: :class:`int`
            The channel_id of the channel to send message to

        message: :class:`str`
            message to send

        is_action: :class:`bool`
            whether the message is an action

        uuid: Optional[:class:`str`]

        **Returns**

        :class:`ChatMessage`
        """
        data = {'is_action': str(is_action).lower(), 'message': message}
        if uuid is not None:
            data["uuid"] = uuid
        data = create_multipart_formdata(data)
        return ChatMessage(self.http.make_request(Path.send_message_to_channel(channel_id), files=data))

    def join_channel(self, channel: str, user: str) -> ChatChannel:
        """
        This endpoint allows you to join a public channel.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        channel: :class:`str`
            channel id of channel to join

        user: :class:`str`
            user joining (you)

        **Returns**

        :class:`ChatChannel`
        """
        return ChatChannel(self.http.make_request(Path.join_channel(channel, user)))

    def leave_channel(self, channel: str, user: str):
        """
        This endpoint allows you to leave a public channel.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        channel: :class:`str`
            channel id of channel to leave

        user: :class:`str`
            user leaving (you)
        """
        self.http.make_request(Path.leave_channel(channel, user))

    def mark_channel_as_read(self, channel_id: int, message_id: int):
        """
        This endpoint marks the channel as having being read up to the given message_id.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        channel_id: :class:`int`
            The channel_id of the channel to mark as read

        message_id: :class:`int`
            The message_id of the message to mark as read up to
        """
        self.http.make_request(Path.mark_channel_as_read(channel_id, message_id))

    def get_channel_list(self) -> Sequence[ChatChannel]:
        """
        This endpoint returns a list of all joinable public channels.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Returns**

        Sequence[:class:`ChatChannel`]
        """
        return list(map(ChatChannel, self.http.make_request(Path.get_channel_list())))

    def create_channel(self, channel_type: Union[ChatChannelType, str], target_id: Optional[int] = None,
                       target_ids: Optional[Sequence[int]] = None, message: Optional[str] = None,
                       channel: Optional[Dict[str, str]] = None) -> ChatChannel:
        """
        [This description may be outdated]

        This endpoint creates a new channel if doesn't exist and joins it.
        Currently only for rejoining existing PM channels which the user has left.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameter**

        channel_type: Union[:class:`ChatChannelType`, :class:`str`]
            channel type (currently only supports "PM" and "ANNOUNCE")

        target_id: Optional[:class:`int`]
            target user id; required if type is PM; ignored, otherwise.

        target_ids: Optional[Sequence[:class:`int`]]
            target user ids; required if type is PM; ignored, otherwise.

        message: Optional[:class:`str`]
            message to send with the announcement; required if type is ANNOUNCE.

        channel: Optional[Dict[str, str]]
            channel details; required if type is ANNOUNCE.

            name: :class:`str`
                the channel name

            description: :class:`str`
                the channel description

        **Returns**

        :class:`ChatChannel`
             contains recent_messages attribute. Note that if there's no existing PM channel,
             most of the fields will be blank. In that case, send a message (create_new_pm)
             instead to create the channel.
        """
        channel_type = parse_enum_args(channel_type)
        if channel_type == "PM":
            if target_id is None and target_ids is None:
                raise ValueError("target_id and target_ids cannot both be null if the channel type is PM")
            data = {"type": channel_type,
                    **({"target_ids": target_ids} if target_ids is not None else {"target_id": target_id})}
        elif channel_type == "ANNOUNCE":
            if message is None:
                raise ValueError("message cannot be null when the channel type is ANNOUNCE")
            elif channel is None or channel["name"] is None or channel["description"] is None:
                raise ValueError("channel and it's items cannot be null when the channel type is ANNOUNCEMENT")
            data = {"type": channel_type, "message": message, "channel": channel}
        else:
            raise ValueError(f"{channel_type} is not a valid channel type that can be created. "
                             f"Check for casing (uppercase) if the type is correct.")
        return ChatChannel(self.http.make_request(Path.create_channel(), data=data))

    def get_channel(self, channel_id: int) -> dict:
        """
        Gets details of a chat channel.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameter**

        channel_id: :class:`int`

        **Returns**

        :class:`dict`
            {
            channel: :class:`ChatChannel`,

            users: :class:`UserCompact`
                Only visible for PM channels

            }
        """
        resp = self.http.make_request(Path.get_channel(channel_id))
        return {
            'channel': ChatChannel(resp['channel']),
            'users': list(map(UserCompact, resp['users'])),
        }

    def get_comments(self, commentable_type: Optional[Union[ObjectType, str]] = None,
                     commentable_id: Optional[int] = None, cursor: Optional[dict] = None,
                     parent_id: Optional[int] = None,
                     sort: Optional[Union[str, CommentSort]] = None) -> CommentBundle:
        """
        Returns a list comments and their replies up to 2 levels deep.

        Does not require OAuth

        **Parameter**

        commentable_type: Optional[Union[:class:`ObjectType`, :class:`str`]
            The type of resource to get comments for. Must be of the following types:
            beatmapset, build, news_post

        commentable_id: Optional[:class:`int`]
            The id of the resource to get comments for. Id correlates with commentable_type.

        cursor: Optional[:class:`dict`]
            Pagination option. See :class:`CommentSort` for detail.
            The format follows Cursor except it's not currently included in the response.

        parent_id: Optional[:class:`int`]
            Limit to comments which are reply to the specified id. Specify 0 to get top level comments.

        sort: Optional[Union[:class:`str`, :class:`CommentSort`]]
            Sort option as defined in :class:`CommentSort`.
            Defaults to new for guests and user-specified default when authenticated.

        **Returns**

        :class:`CommentBundle`
            pinned_comments is only included when commentable_type and commentable_id are specified.
        """
        commentable_type, sort = parse_enum_args(commentable_type, sort)
        if commentable_type is not None and commentable_type not in ("beatmapset", "build", "news_post"):
            raise ValueError("commentable_type, if not null, must be of the following: beatmapset, build, new_post")
        return CommentBundle(self.http.make_request(Path.get_comments(), commentable_type=commentable_type,
                                                    commentable_id=commentable_id, **(cursor if cursor else {}),
                                                    parent_id=parent_id, sort=sort))

    def post_comment(self, commentable_type: Optional[Union[ObjectType, str]] = None,
                     commentable_id: Optional[int] = None, message: Optional[str] = None,
                     parent_id: Optional[int] = None) -> CommentBundle:
        """
        Posts a new comment to a comment thread.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameter**

        commentable_type: Optional[Union[:class:`ObjectType`, :class:`str`]
            The type of resource to get comments for. Must be of the following types:
            beatmapset, build, news_post

        commentable_id: Optional[:class:`int`]
            The id of the resource to get comments for. Id correlates with commentable_type.

        message: Optional[:class:`str`]
            Text of the comment

        parent_id: Optional[:class:`int`]
            The id of the comment to reply to, null if not a reply

        **Returns**

        :class:`CommentBundle`
        """
        params = {
            "comment[commentable_type]": parse_enum_args(commentable_type),
            "comment[commentable_id]": commentable_id,
            "comment[message]": message,
            "comment[parent_id]": parent_id
        }
        return CommentBundle(self.http.make_request(Path.post_new_comment(), **params))

    def get_comment(self, comment: int) -> CommentBundle:
        """
        Gets a comment and its replies up to 2 levels deep.

        Does not require OAuth

        **Parameters**

        comment: :class:`int`
            Comment id

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(self.http.make_request(Path.get_comment(comment)))

    def edit_comment(self, comment: int, message: Optional[str] = None) -> CommentBundle:
        """
        Edit an existing comment.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        comment: :class:`int`
            Comment id

        message: Optional[:class:`str`]
            New text of the comment

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(self.http.make_request(Path.edit_comment(comment),
                                                    **{"comment[message]": message}))

    def delete_comment(self, comment: int) -> CommentBundle:
        """
        Deletes the specified comment.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        comment: :class:`int`
            Comment id

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(self.http.make_request(Path.delete_comment(comment)))

    def add_comment_vote(self, comment: int) -> CommentBundle:
        """
        Upvotes a comment.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        comment: :class:`int`
            Comment id

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(self.http.make_request(Path.add_comment_vote(comment)))

    def remove_comment_vote(self, comment: int) -> CommentBundle:
        """
        Un-upvotes a comment.

        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        comment: :class:`int`
            Comment id

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(self.http.make_request(Path.remove_comment_vote(comment)))

    def reply_topic(self, topic: int, body: str) -> ForumPost:
        """
        Create a post replying to the specified topic.

        Requires OAuth, scope forum.write, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        topic: :class:`int`
            Id of the topic to be replied to.

        body: :class:`str`
            Content of the reply post.

        **Returns**

        :class:`ForumPost`
            body attributes included
        """
        data = {'body': body}
        return ForumPost(self.http.make_request(Path.reply_topic(topic), data=data))

    def create_topic(self, body: str, forum_id: int, title: str, with_poll: Optional[bool] = None,
                     hide_results: Optional[bool] = None, length_days: Optional[int] = None,
                     max_options: Optional[int] = None, poll_options: Optional[str] = None,
                     poll_title: Optional[str] = None, vote_change: Optional[bool] = None) -> dict:
        """
        Create a new topic.

        Requires OAuth, scope forum.write, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        body: :class:`str`
            Content of the topic.

        forum_id: :class:`int`
            Forum to create the topic in.

        title: :class:`str`
            Title of the topic.

        with_poll: Optional[:class:`bool`]
            Enable this to also create poll in the topic (default: false).

        hide_results: Optional[:class:`bool`]
            Enable this to hide result until voting period ends (default: false).

        length_days: Optional[:class:`int`]
            Number of days for voting period. 0 means the voting will never ends (default: 0).
            This parameter is required if hide_results option is enabled.

        max_options: Optional[:class:`int`]
            Maximum number of votes each user can cast (default: 1).

        poll_options: Optional[:class:`str`]
            Newline-separated list of voting options. BBCode is supported.

        poll_title: Optional[:class:`str`]
            Title of the poll.

        vote_change: Optional[:class:`bool`]
            Enable this to allow user to change their votes (default: false).

        **Returns**

        :class:`dict`
            {
            topic: :class:`ForumTopic`

            post: :class:`ForumPost`
                includes body

            }
        """
        data = {'body': body, 'forum_id': forum_id, 'title': title, 'with_poll': with_poll}
        if with_poll:
            if poll_options is None or poll_title is None:
                raise TypeError("poll_options and poll_title are required since the topic has a poll.")
            data.update({'forum_topic_poll': {
                'hide_results': hide_results, 'length_days': length_days,
                'max_options': max_options, 'poll_options': poll_options,
                'poll_title': poll_title, 'vote_change': vote_change
            }})
        resp = self.http.make_request(Path.create_topic(), data=json.dumps(data))
        return {
            'topic': ForumTopic(resp['topic']),
            'post': ForumPost(resp['post'])
        }

    def get_topic_and_posts(self, topic: int, cursor: Optional[dict] = None, sort: Optional[str] = None,
                            limit: Optional[int] = None, start: Optional[int] = None,
                            end: Optional[int] = None) -> dict:
        """
        Get topic and its posts.

        Requires OAuth and scope public

        **Parameters**

        topic: :class:`int`
            Id of the topic.

        cursor: Optional[:class:`dict`]
            To be used to fetch the next page of results

        sort: Optional[:class:`str`]
            Post sorting option. Valid values are id_asc (default) and id_desc.

        limit: Optional[:class:`int`]
            Maximum number of posts to be returned (20 default, 50 at most).

        start: Optional[:class:`int`]
            First post id to be returned with sort set to id_asc.
            This parameter is ignored if cursor is specified.

        end: Optional[:class:`int`]
            First post id to be returned with sort set to id_desc.
            This parameter is ignored if cursor is specified.

        **Returns**

        :class:`dict`
            {
            cursor: :class:`dict`,

            search: :class:`dict`,

            posts: Sequence[:class:`ForumPost`],

            topic: :class:`ForumTopic`

            }
        """
        resp = self.http.make_request(Path.get_topic_and_posts(topic), **(cursor if cursor else {}),
                                      sort=sort, limit=limit, start=start, end=end)
        return {
            'cursor': resp['cursor'],
            'search': resp['search'],
            'posts': list(map(ForumPost, resp['posts'])),
            'topic': ForumTopic(resp['topic'])
        }

    def edit_topic(self, topic: int, topic_title: str) -> ForumTopic:
        """
        Edit topic. Only title can be edited through this endpoint.

        Requires OAuth and scope forum.write

        **Parameters**

        topic: :class:`int`
            Id of the topic.

        topic_title: :class:`str`
            New topic title.

        **Returns**

        :class:`ForumTopic`
        """
        data = {'forum_topic': {'topic_title': topic_title}}
        return ForumTopic(self.http.make_request(Path.edit_topic(topic), data=json.dumps(data)))

    def edit_post(self, post: int, body: str) -> ForumPost:
        """
        Edit specified forum post.

        Requires OAuth and scope forum.write

        post: :class:`int`
            Id of the post.

        body: :class:`str`
            New post content in BBCode format.

        **Returns**

        :class:`ForumPost`
        """
        data = {'body': body}
        return ForumPost(self.http.make_request(Path.edit_post(post), data=data))

    def search(self, mode: Optional[Union[str, WikiSearchMode]] = None, query: Optional[str] = None,
               page: Optional[int] = None) -> Dict[str, SearchResults]:
        """
        Searches users and wiki pages.

        Requires OAuth and scope public

        **Parameters**

        mode: Optional[Union[:class:`str`, :class:`WikiSearchMode`]]
            Either all, user, or wiki_page. Default is all.

        query: Optional[:class:`str`]
            Search keyword.

        page: Optional[:class:`int`]
            Search result page. Ignored for mode all.

        **Returns**

        Dict[:class:`str`, :class:`SearchResults`]
            {

            "user": Union[:class:`SearchResults`, :class:`None`]

            "wiki_page": Union[:class:`SearchResults`, :class:`None`]

            }
        """
        mode = parse_enum_args(mode)
        resp = self.http.make_request(Path.search(), mode=mode, query=query, page=page)
        return {
            'user': SearchResults(resp["user"], UserCompact) if 'user' in resp else None,
            'wiki_page': SearchResults(resp["wiki_page"], WikiPage) if 'wiki_page' in resp else None,
        }

    def get_user_highscore(self, room: int, playlist: int, user: int) -> MultiplayerScore:
        """
        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        room: :class:`int`
            Id of the room.

        playlist: :class:`int`
            Id of the playlist item.

        user: :class:`int`
            User id.

        **Returns**

        :class:`MultiplayerScores`
        """
        return MultiplayerScore(self.http.make_request(Path.get_user_high_score(room, playlist, user)))

    def get_scores(self, room: int, playlist: int, limit: Optional[int] = None,
                   sort: Optional[Union[str, MultiplayerScoresSort]] = None,
                   cursor: Optional[dict] = None) -> MultiplayerScores:
        """
        Requires OAuth, scope public, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        room: :class:`int`
            Id of the room.

        playlist: :class:`int`
            Id of the playlist item.

        limit: Optional[:class:`int`]
            Number of scores to be returned.

        sort: Optional[Union[:class:`str`, :class:`MultiplayerScoresSort`]]


        cursor: Optional[:class:`dict`]

        **Returns**

        :class:`MultiplayerScores`
        """
        sort = parse_enum_args(sort)
        return MultiplayerScores(self.http.make_request(Path.get_scores(room, playlist),
                                                        limit=limit, sort=sort, **(cursor if cursor else {})))

    def get_score(self, room: int, playlist: int, score: int) -> MultiplayerScore:
        """
        Requires OAuth, scope lazer, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        room: :class:`int`
            Id of the room.

        playlist: :class:`int`
            Id of the playlist item.

        score: :class:`int`
            Id of the score.

        **Returns**

        :class:`MultiplayerScore`
        """
        return MultiplayerScore(self.http.make_request(Path.get_score(room, playlist, score)))

    def get_news_listing(self, limit: Optional[int] = None, year: Optional[int] = None,
                         cursor: Optional[dict] = None) -> dict:
        """
        Returns a list of news posts and related metadata.

        **Parameters**

        limit: Optional[:class:`int`]
            Maximum number of posts (12 default, 1 minimum, 21 maximum).

        year: Optional[:class:`int`]
            Year to return posts from.

        cursor: Optional[:class:`dict`]
            Cursor for pagination.

        **Returns**

        {

        cursor: :class:`dict`

        news_posts: Sequence[:class:`NewsPost`]
            Includes preview.

        news_sidebar: {

            current_year: :class:`int`
                Year of the first post's publish time, or current year if no posts returned.

            years: :class:`int`
                All years during which posts have been published.

            news_posts: Sequence[:class:`NewsPost`]
                All posts published during current_year.

        }

        search: {

            limit: :class:`int`
                Clamped limit input.

            sort: :class:`str`
                Always published_desc.

            }

        }
        """
        response = self.http.make_request(Path.get_news_listing(), limit=limit, year=year, cursor=cursor)
        return {
            "cursor": response['cursor'],
            "news_posts": list(map(NewsPost, response["news_posts"])),
            "news_sidebar": {
                "current_year": response['news_sidebar']['current_year'],
                "years": response['news_sidebar']['years'],
                "news_posts": list(map(NewsPost, response['news_sidebar']['news_posts'])),
            },
            "search": response['search']
        }

    def get_news_post(self, news: str, key: Optional[str] = None) -> NewsPost:
        """
        Returns details of the specified news post.

        **Parameters**

        news: class:`str`
            News post slug or ID.

        key: Optional[:class:`str`]
            Unset to query by slug, or id to query by ID.

        **Returns**

        Returns a :class:`NewsPost` with content and navigation included.
        """
        return NewsPost(self.http.make_request(Path.get_news_post(news), key=key))

    def get_notifications(self, max_id: Optional[int] = None) -> dict:
        """
        This endpoint returns a list of the user's unread notifications. Sorted descending by id with limit of 50.

        Requires OAuth, scope lazer, a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        max_id: Optional[:class:`int`]
            Maximum id fetched. Can be used to load earlier notifications.
            Defaults to no limit (fetch latest notifications)

        **Returns**

        :class:`dict`
            {

            has_more: :class:`bool`,
                whether or not there are more notifications

            notifications: Sequence[:class:`Notification`],

            unread_count: :class:`bool`
                total unread notifications

            notification_endpoint: :class:`str`
                url to connect to websocket server

            }
        """
        resp = self.http.make_request(Path.get_notifications(), max_id=max_id)
        return {
            "notifications": list(map(Notification, resp['notifications'])),
            "stacks": resp["stacks"],
            "timestamp": resp["timestamp"],
            "types": resp["types"],
            "notification_endpoint": resp['notification_endpoint'],
        }

    def mark_notifications_read(self, identities: Optional[Sequence[Union[
        IdentitiesUtil, Dict[str, Union[str, int]]]]] = None,
        notifications: Optional[Sequence[Union[
            NotificationsUtil, Dict[str, str]]]] = None):
        """
        This endpoint allows you to mark notifications read. Should only supply one of the arguments.

        Requires OAuth, scope lazer, a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        identities: Sequence[Union[:class:`ReadNotifIdentitiesUtil`,
        Dict[:class:`str`, Union[:class:`str`, :class:`int`]]]]

        notifications: Sequence[Union[:class:`ReadNotifNotificationsUtil`, Dict[:class:`str`, :class:`str`]]]
        """
        if identities is not None and notifications is not None:
            raise ValueError("Should only supply one argument, either identities or notifications.")
        elif identities is None and notifications is None:
            raise ValueError("Must supply at least one argument, either identities or notifications.")
        try:
            name = "identities" if notifications is None else "notifications"
            params = JsonUtil.list_to_labeled_dict(locals()[name], name)
        except Exception as exc:
            raise ValueError("An error occurred while parsing the argument you provided.") from exc
        self.http.make_request(Path.mark_notifications_as_read(), **params)

    def revoke_current_token(self):
        """
        Revokes currently authenticated token.

        Requires OAuth
        """
        self.http.make_request(Path.revoke_current_token())

    def get_ranking(self, mode: Union[str, GameModeStr], type: Union[str, RankingType],
                    country: Optional[str] = None, cursor: Optional[dict] = None,
                    filter: Optional[str] = None, spotlight: Optional[int] = None,
                    variant: Optional[str] = None) -> Rankings:
        """
        Gets the current ranking for the specified type and game mode.

        Requires OAuth and scope public

        mode: Union[:class:`str`, :class:`GameModeStr`]

        type: Union[:class:`str`, :class:`RankingType`]
            :class:`RankingType`

        country: Optional[:class:`str`]
            Filter ranking by country code. Only available for type of performance.

        cursor: Optional[:class:`dict`]

        filter: Optional[:class:`str`]
            Either all (default) or friends.

        spotlight: Optional[:class:`int`]
            The id of the spotlight if type is charts.
            Ranking for latest spotlight will be returned if not specified.

        variant: Optional[:class:`str`]
            Filter ranking to specified mode variant.
            For mode of mania, it's either 4k or 7k. Only available for type of performance.

        **Returns**

        :class:`Rankings`
        """
        mode, type = parse_enum_args(mode, type)
        return Rankings(self.http.make_request(Path.get_ranking(mode, type), country=country,
                                               **(cursor if cursor else {}), filter=filter,
                                               spotlight=spotlight, variant=variant))

    def get_spotlights(self) -> Spotlights:
        """
        Gets the list of spotlights.

        Requires OAuth and scope public

        **Returns**

        :class:`Spotlights`
        """
        return Spotlights(self.http.make_request(Path.get_spotlights()))

    def get_own_data(self, mode: Union[str, GameModeStr] = "") -> User:
        """
        Similar to get_user but with authenticated user (token owner) as user id.

        Requires OAuth, scope identify, and a user (authorization code grant, delegate scope, or password auth)

        **Parameters**

        mode: Optional[:class:`str`, :class:`GameModeStr`]
            GameMode. User default mode will be used if not specified.

        **Returns**

        See return for get_user
        """
        mode = parse_enum_args(mode)
        return User(self.http.make_request(Path.get_own_data(mode)))

    def get_user_kudosu(self, user: int, limit: Optional[int] = None, offset: Optional[int] = None):
        """
        Returns kudosu history.

        Requires OAuth and scope public

        **Parameters**

        user: :class:`int`
            Id of the user.

        limit: Optional[:class:`int`]
            Maximum number of results.

        offset: Optional[:class:`int`]
            Result offset for pagination.

        **Returns**

        Sequence[:class:`KudosuHistory`]
        """
        return list(map(KudosuHistory, self.http.make_request(Path.get_user_kudosu(user),
                                                              limit=limit, offset=offset)))

    def get_user_scores(self, user: int, type: Union[UserScoreType, str], include_fails: Optional[int] = None,
                        mode: Optional[Union[str, GameModeStr]] = None, limit: Optional[int] = None,
                        offset: Optional[int] = None) -> Sequence[LegacyScore]:
        """
        This endpoint returns the scores of specified user.

        Requires OAuth and scope public

        **Parameters**

        user: :class:`int`
            Id of the user.

        type: union[:class:`UserScoreType` :class:`str`]
            Score type. Must be one of these: best, firsts, recent

        include_fails: Optional[:class:`int`]
            Only for recent scores, include scores of failed plays. Set to 1 to include them. Defaults to 0.

        mode: Optional[Union[:class:`GameModeStr`, :class:`str`]]
            game mode of the scores to be returned. Defaults to the specified user's mode.

        limit: Optional[:class:`int`]
            Maximum number of results.

        offset: Optional[:class:`int`]
            Result offset for pagination.

        **Returns**

        Sequence[:class:`LegacyScore`]
            Includes attributes beatmap, beatmapset, weight: Only for type best, user
        """
        mode, type = parse_enum_args(mode, type)
        return [LegacyScore(score) for score in self.http.make_request(Path.get_user_scores(user, type),
                                                                       include_fails=include_fails, mode=mode,
                                                                       limit=limit, offset=offset)]

    def get_user_beatmaps(self, user: int, type: Union[str, UserBeatmapType], limit: Optional[int] = None,
                          offset: Optional[int] = None) -> Sequence[Union[BeatmapPlaycount, Beatmapset]]:
        """
        Returns the beatmaps of specified user.

        Requires OAuth and scope public

        **Parameters**

        user: :class:`int`
            Id of the user.

        type: Union[:class:`str`, :class:`UserBeatmapType`]
            Beatmap type. Can be one of the following - favourite, graveyard, loved, most_played, pending, ranked.

        limit: Optional[:class:`int`]
            Maximum number of results.

        offset: Optional[:class:`int`]
            Result offset for pagination.

        **Returns**

        Sequence[Union[:class:`BeatmapPlaycount`, :class:`Beatmapset`]]
            :class:`BeatmapPlaycount` for type most_played or :class:`Beatmapset` for any other type.
        """
        object_type = Beatmapset
        type = parse_enum_args(type)
        if type == 'most_played':
            object_type = BeatmapPlaycount
        return list(map(object_type, self.http.make_request(Path.get_user_beatmaps(user, type),
                                                            limit=limit, offset=offset)))

    def get_user_recent_activity(self, user: int, limit: Optional[int] = None,
                                 offset: Optional[int] = None) -> Sequence[Event]:
        """
        Returns recent activity.

        Requires OAuth and scope public

        **Parameters**

        user: :class:`int`
            Id of the user.

        limit: Optional[:class:`int`]
            Maximum number of results.

        offset: Optional[:class:`int`]
            Result offset for pagination.

        **Returns**

        Sequence[:class:`Event`]
            list of :class:`Event` objects
        """
        return list(map(Event, self.http.make_request(Path.get_user_recent_activity(user),
                                                      limit=limit, offset=offset)))

    def get_user(self, user: int, mode: Optional[Union[str, GameModeStr]] = '', key: Optional[str] = None) -> User:
        """
        This endpoint returns the detail of specified user.

        Requires OAuth and scope public

        **Parameters**

        user: Union[:class:`int`, :class:`str`]
            Id or username of the user. Id lookup is prioritised unless key parameter is specified.
            Previous usernames are also checked in some cases.

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]
            User default mode will be used if not specified.

        key: Optional[:class:`str`]
            Type of user passed in url parameter. Can be either id or username
            to limit lookup by their respective type. Passing empty or invalid
            value will result in id lookup followed by username lookup if not found.

        **Returns**

        :class:`User`
            Includes following attributes: account_history, active_tournament_banner,
            badges, beatmap_playcounts_count, favourite_beatmapset_count, follower_count,
            graveyard_beatmapset_count, groups, loved_beatmapset_count,
            mapping_follower_count, monthly_playcounts, page, pending_beatmapset_count,
            previous_usernames, rank_history, ranked_beatmapset_count, replays_watched_counts,
            scores_best_count, scores_first_count, scores_recent_count, statistics,
            statistics.country_rank, statistics.rank, statistics.variants, support_level,
            user_achievements.
        """
        mode = parse_enum_args(mode)
        return User(self.http.make_request(Path.get_user(user, mode), key=key))

    def get_users(self, ids: Sequence[int]) -> Sequence[UserCompact]:
        """
        Returns list of users.

        Requires OAuth and scope public

        **Parameters**

        ids: Sequence[:class:`int`]
            User id to be returned. Specify once for each user id requested.
            Up to 50 users can be requested at once.

        **Returns**

        Sequence[:class:`UserCompact`]
            Includes attributes: country, cover, groups, statistics_rulesets.
        """
        res = self.http.make_request(Path.get_users(), **{"ids[]": ids})
        return list(map(UserCompact, res["users"]))

    def get_wiki_page(self, locale: str, path: str) -> WikiPage:
        """
        The wiki article or image data.

        No OAuth required.

        **Parameters**

        locale: :class:`str`
            Two-letter language code of the wiki page.

        path: :class:`str`
            The path name of the wiki page.

        **Returns**

        :class:`WikiPage`
        """
        return WikiPage(self.http.make_request(Path.get_wiki_page(locale, path)))

    def get_beatmapset_events(self, page: Optional[int] = None, limit: Optional[int] = None,
                              sort: Optional[Union[str, BeatmapsetEventSort]] = None,
                              type: Optional[Union[str, BeatmapsetEventType]] = None,
                              min_date: Optional[Union[str, datetime]] = None,
                              max_date: Optional[Union[str, datetime]] = None) -> \
            Dict[str, Union[Sequence[BeatmapsetEvent], Dict, Sequence[UserCompact]]]:
        """
        Returns a list of beatmapset events.

        Requires OAuth and scope public.

        **Parameters**

        page: Optional[:class:`int`]

        limit: Optional[:class:`int`]

        sort: Optional[Union[:class:`str`, :class:`BeatmapsetEventSort`]]
            Specified a sort order.

        type: Optional[Union[:class:`str`, :class:`BeatmapsetEventType`]]
            Specifies for only a certain type of event to be returned.

        **Returns**

        Dict[str, Union[Sequence[BeatmapsetEvent], Dict, Sequence[UserCompact]]]

        {

            'events': Sequence[BeatmapsetEvent],

            'reviews_config': Dict,

            'users': Sequence[UserCompact]

        }
        """
        sort, type = parse_enum_args(sort, type)
        if isinstance(min_date, datetime):
            min_date = min_date.isoformat()
        if isinstance(max_date, datetime):
            max_date = max_date.isoformat()
        resp = self.http.make_request(Path.get_beatmapset_events(), page=page, limit=limit, sort=sort,
                                      type=type, min_date=min_date, max_date=max_date)
        return {
            "events": [BeatmapsetEvent(event) for event in resp['events']],
            "reviews_config": resp['reviewsConfig'],
            "users": [UserCompact(user) for user in resp['users']],
        }

    def get_matches(self, limit: Optional[int] = None, sort: Optional[Union[str, MatchSort]] = None) \
            -> Dict[str, Union[Sequence[Match], Dict]]:
        """
        Returns a list of matches.

        Requires OAuth and scope public.

        **Parameters**

        limit: Optional[:class:`int`]

        sort: Optional[Union[:class:`str`, :class:`MatchSort`]]

        **Returns**

        Dict[:class:`str`, Union[Dict, Sequence[:class:`Match`]]]
        """
        sort = parse_enum_args(sort)
        resp = self.http.make_request(Path.get_matches(), limit=limit, sort=sort)
        return {
            "matches": list(map(Match, resp['matches'])),
            "cursor": resp['cursor'],
            "params": resp['params'],
        }

    def get_match(self, match_id: int) -> Match:
        """
        Returns a match by id.

        Requires OAuth and scope public.

        **Parameters**

        match_id: :class:`int`
            The match id.

        **Returns**

        :class:`Match`
        """
        return MatchExtended(self.http.make_request(Path.get_match(match_id)))

    def get_rooms(self, mode: Union[str, GameModeStr] = '', sort: Optional[Union[str, RoomSort]] = None,
                  limit: Optional[int] = None,
                  room_type: Optional[Union[RoomType, str]] = None,
                  category: Optional[Union[RoomCategory, str]] = None,
                  filter_mode: Optional[Union[RoomFilterMode, str]] = None) -> Sequence[Room]:
        """
        Returns a list of rooms.

        Requires OAuth, scope public, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            Game mode to filter rooms by.

        sort: Optional[Union[:class:`str`, :class:`RoomSort`]]
            Sort rooms by.

        limit: Optional[:class:`int`]
            max number of rooms to return

        room_type: Optional[Union[:class:`RoomType`, :class:`str`]]
            type of room to look for

        category: Optional[Union[:class:`RoomCategory`, :class:`str`]]
            type of category of room to look for

        filter_mode: Optional[Union[:class:`RoomFilterMode`, :class:`str`]]
        """
        mode, sort, room_type, category, filter_mode = parse_enum_args(
            mode, sort, room_type, category, filter_mode)
        return list(map(Room, self.http.make_request(Path.get_rooms(mode), sort=sort, limit=limit,
                                                     type_group=room_type, category=category, mode=filter_mode)))

    def get_seasonal_backgrounds(self) -> SeasonalBackgrounds:
        """
        Get the season backgrounds.

        Doesn't require OAuth

        **Returns**

        :class:`SeasonalBackgrounds`
        """
        return SeasonalBackgrounds(self.http.make_request(Path.get_seasonal_backgrounds()))

    def get_room(self, room_id: int) -> Room:
        """
        Returns a room by id.

        Requires OAuth and scope public.

        **Parameters**

        room_id: :class:`int`
            The room id.

        **Returns**

        :class:`Room`
        """
        return Room(self.http.make_request(Path.get_room(room_id)))

    def get_score_by_id(self, mode, score_id) -> LegacyScore:
        """
        Returns a score by id.

        Requires OAuth and scope public.

        **Parameters**

        mode: Union[:class:`str`, :class:`GameModeStr`]

        score_id: :class:`int`

        **Returns**

        :class:`LegacyScore`
        """
        mode = parse_enum_args(mode)
        return LegacyScore(self.http.make_request(Path.get_score_by_id(mode, score_id)))

    def search_beatmapsets(self, filters=None, page=None) -> dict:
        """
        Search for beatmapsets.

        Requires OAuth and scope public.

        **Attributes**

        filters: Optional[:class:`BeatmapsetSearchFilter`]

        page: Optional[:class:`int`]

        **Returns**

        {

        "beatmapsets": Sequence[:class:`Beatmapset`]

        "cursor": :class:`dict`,

        "search": :class:`dict`,

        "recommended_difficulty": Union[:class:`float`, :class:`None`]

        "error": Union[:class:`str`, :class:`None`]

        "total": :class:`int`

        }
        """
        if filters is None:
            filters = {}
        if isinstance(filters, BeatmapsetSearchFilter):
            filters = filters.filters
        resp = self.http.make_request(Path.beatmapset_search(), page=page, **filters)
        return {
            'beatmapsets': [Beatmapset(beatmapset) for beatmapset in resp['beatmapsets']],
            'cursor': resp['cursor'],
            'search': resp['search'],
            'recommended_difficulty': resp['recommended_difficulty'],
            'error': resp['error'],
            'total': resp['total'],
        }

    def get_room_leaderboard(self, room_id: int) -> \
            Dict[str, Union[Sequence[UserScoreAggregate], Union[UserScoreAggregate, None]]]:
        """
        Return a room's leaderboard. The :class:`UserScoreAggregate` objects returned under the "leaderboard"
        key contain the "user" attribute. The :class:`UserScoreAggregate` object under the "user_score" key
        contains the "user" and "position" attributes.

        Requires OAuth, scope public, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        room_id: :class:`int`

        **Returns**

        Dict[:class:`str`, Union[Sequence[:class:`UserScoreAggregate`],
        Union[:class:`UserScoreAggregate`, :class:`NoneType`]]]

        {

            'leaderboard': Sequence[:class:`UserScoreAggregate`],

            'user_score': Union[:class:`UserScoreAggregate`, :class:`NoneType`]

        }
        """
        resp = self.http.make_request(Path.get_room_leaderboard(room_id))
        return {
            'leaderboard': list(map(UserScoreAggregate, resp['leaderboard'])),
            'user_score': UserScoreAggregate(resp['user_score']) if resp['user_score'] is not None else None,
        }

    def get_replay_data(self, mode, score_id):
        """
        Returns replay data for a score.

        Requires OAuth, scope public, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        mode: Union[:class:`str`, :class:`GameModeStr`]

        score_id: :class:`int`

        **Returns**

        :class:`osrparse.Replay`
        """
        mode = parse_enum_args(mode)
        return Replay.from_string(self.http.make_request(Path.get_replay_data(mode, score_id), is_download=True))

    def get_friends(self):
        """
        Returns a list of friends.

        Requires OAuth, scope friends.read, and a user (authorization code grant, delegate scope, or password auth).

        **Returns**

        Sequence[:class:`User`]
        """
        return list(map(UserCompact, self.http.make_request(Path.get_friends())))

    def get_new_score_token(self, beatmap_id: int, version_hash: str, beatmap_hash: str,
                            ruleset: Union[GameModeInt, GameModeStr, int]) -> dict:
        """
        Creates a score token with which can be used to submit a score.

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        beatmap_id: :class:`int`
            id of the beatmap

        version_hash: :class:`str`
            version hash is the md5 checksum of f"{game_version}{os_enum}"
            os_enum is Windows = 1, Linux = 2, macOS = 3, iOS = 4, Android = 5

        beatmap_hash: :class:`str`
            md5 hash of the beatmap

        ruleset: Union[:class:`GameModeStr`, :class:`GameModeInt`, :class:`int`]

        **Returns**

        :class:`dict`

        {

        "beatmap_id": :class:`int`,

        "created_at": :class:`datetime.datetime`

        "id": :class:`int`

        "user_id": :class:`int`

        }
        """
        if isinstance(ruleset, GameModeStr):
            ruleset = GameModeStr.get_int_equivalent(ruleset)
        ruleset = parse_enum_args(ruleset)
        data = {"version_hash": version_hash, "beatmap_hash": beatmap_hash, "ruleset_id": ruleset}
        resp = self.http.make_request(Path.get_new_score_id(beatmap_id), files=create_multipart_formdata(data))
        resp["created_at"] = parser.parse(resp["created_at"])
        return resp

    def submit_score(self, beatmap_id, token, score_data) -> SoloScore:
        """
        Submits a score

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        beatmap_id: :class:`int`
            id of the beatmap

        token: :class:`str`
            score token which can be obtained from Client.get_new_score_token

        score_data: :class:`dict`
            data of the score which is being submitted
        """
        return SoloScore(self.http.make_request(Path.submit_score(beatmap_id, token), data=score_data))

    def favourite_beatmapset(self, beatmapset_id: int, favourite: bool) -> int:
        """
        Add or remove a favourite beatmapset

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        beatmapset_id: :class:`int`

        favourite: :class:`bool`
            whether to favourite (true) or unfavourite (false)

        **Returns**

        :class:`int`
            The number of favourites on the beatmapsets
        """
        resp = self.http.make_request(Path.favourite_beatmapset(beatmapset_id),
                                      data={"action": "favourite" if favourite else "unfavourite"})
        return resp["favourite_count"]

    def get_open_chat_channels(self):
        """
        Get a list of chat channels that you have open. Includes recent DMs and public chat channels.

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        **Returns**

        Sequence[:class:`ChatChannel`]
        """
        return list(map(ChatChannel, self.http.make_request(Path.get_chat_presence())))

    def join_user_to_room(self, room: int, user: int, password: Optional[str] = None):
        """
        Invite a user to a room.

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        room: :class:`int`

        user: :class:`int`

        password: Optional[:class:`str`]
        """
        return self.http.make_request(Path.join_to_room(room, user), password=password)

    def kick_user_from_room(self, room: int, user: int):
        """
        Kick a user from a room (or leave by kicking yourself).

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        room: :class:`int`

        user: :class:`int`
        """
        return self.http.make_request(Path.kick_from_room(room, user))

    def report(self, comments: str, reason: str, reportable_id: int, reportable_type: Union[ObjectType, str]):
        """
        Send a report.

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        comments: :class:`str`

        reason: :class:`str`

        reportable_id: :class:`id`

        reportable_type: Union[:class:`str`, :class:`ObjectType`]
        """
        params = {
            "comments": comments, "reason": reason, "reportable_id": reportable_id,
            "reportable_type": parse_enum_args(reportable_type)
        }
        self.http.make_request(Path.send_report(), **params)

    def create_multiplayer_room(self, name: str, starting_map: Union[PlaylistItemUtil, dict],
                                password: Optional[str] = None,
                                queue_mode: Optional[Union[RealTimeQueueMode, str]] = RealTimeQueueMode.HOST_ONLY,
                                auto_start_duration: Optional[int] = 0,
                                room_type: Optional[Union[RealTimeType, str]] = RealTimeType.HEAD_TO_HEAD,
                                auto_skip: Optional[bool] = False) -> Room:
        """
        Create a multiplayer (realtime) room.

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        name: :class:`str`
            Name of the room

        starting_map: Union[:class:`PlaylistItemUtil`, dict]

        password: Optional[:class:`str`]
            Password to enter the room which is optional.

        queue_mode: Optional[Union[:class:`RealTimeQueueMode`, :class:`str`]]
            The mode for queuing maps.

        auto_start_duration: Optional[:class:`int`]

        room_type: Optional[Union[:class:`RealTimeType`, :class:`str`]]

        auto_skip: Optional[:class:`bool`]
            Whether to automatically skip intro or not.

        **Returns**

        :class:`Room`
        """
        if isinstance(starting_map, PlaylistItemUtil):
            starting_map = starting_map.json
        queue_mode, room_type = parse_enum_args(queue_mode, room_type)
        data = {
            "name": name, "password": password, "playlist": [starting_map],
            "queue_mode": queue_mode, "auto_start_duration": auto_start_duration,
            "category": "realtime", "type": room_type, "auto_skip": auto_skip,
        }
        return Room(self.http.make_request(Path.create_room(), data=json.dumps(data)))

    def create_playlist(self, name: str, playlist_items: Sequence[Union[PlaylistItemUtil, dict]],
                        duration: Optional[int] = None, ends_at: Optional[Union[str, datetime]] = None,
                        max_attempts: Optional[int] = None,
                        queue_mode: Optional[Union[PlaylistQueueMode, str]] = PlaylistQueueMode.HOST_ONLY,
                        auto_start_duration: Optional[int] = 0):
        """
        Create a playlist

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        **Parameters**

        name: :class:`str`
            Name of the playlist

        playlist_items: Sequence[Union[:class:`PlaylistItemUtil`, :class:`dict`]]
            List of beatmaps to put on the playlist

        duration: Optional[:class:`int`]
            Duration for the playlist to last in minutes.
            If not specified then an end time must be specified.
            The playlist must have a duration of at least 30 minutes.

        ends_at: Optional[Union[:class:`datetime.datetime`, :class:`str`]]
            Time for the playlist to end at. If not specified then a duration must be specified.
            Must amount to a duration of at least 30 minutes.

        max_attempts: Optional[:class:`int`]
            Null means infinite attempts.

        queue_mode: Optional[Union[:class:`PlaylistQueueMode`, :class:`str`]]
            PlaylistQueueMode.HOST_ONLY is the only option

        auto_start_duration: Optional[:class:`int`]
        """
        if duration is None and ends_at is None:
            raise ValueError("Either duration or ends_at must be not null.")
        if ends_at is not None and isinstance(ends_at, datetime):
            ends_at = ends_at.isoformat()
        playlist_items = [item.json if isinstance(item, PlaylistItemUtil) else item for item in playlist_items]
        data = {
            "name": name, "max_attempts": max_attempts, "duration": duration,
            "ends_at": ends_at, "queue_mode": parse_enum_args(queue_mode),
            "auto_start_duration": auto_start_duration, "category": "playlists",
            "playlist": playlist_items,
        }
        return Room(self.http.make_request(Path.create_room(), data=json.dumps(data)))

    def check_download_quota(self):
        """
        Get the amount of quota you've used.

        Requires OAuth, lazer scope, and a user (authorization code grant, delegate scope, or password auth).

        Requires

        **Returns**

        :class:`int`
        """
        return self.http.make_request(Path.download_quota_check())["quota_used"]



from .enums import (
    Mods, Enum, BeatmapsetSearchSort,
    BeatmapsetSearchGeneral, BeatmapsetSearchPlayed,
    BeatmapsetSearchStatus, BeatmapsetSearchExtra,
    ScoreRank, GameModeInt, Mod, NotificationCategory,
    ObjectType
)
from typing import Sequence, Union


def check_scope(func):
    def check(self, other):
        if not isinstance(other, self.__class__) and type(other) != str:
            raise TypeError(f"Cannot compare type Scope with type {type(other).__name__}")
        if isinstance(other, self.__class__):
            other = other.scope
        if other not in self.valid_scopes:
            raise NameError(f"{other} is not a valid scope.")
        return func(self, other)
    return check


def create_autoclass_for_sphinx():
    with open('objects.py', 'r') as f:
        info = f.readline()
        while info:
            if info.startswith('class'):
                name = info.split()[1]
                if '(' in name:
                    name = name.split('(')[0]
                else:
                    name = name[:-1]
                content = f"{name}\n{'^' * len(name)}\n.. autoclass:: osu.{name}\n   :members:\n\n"
                with open('text.txt', 'a') as f2:
                    f2.write(content)
                    f2.close()
            info = f.readline()
        f.close()


def parse_mods_arg(mods):
    if mods is None:
        return
    if isinstance(mods, Mods):
        return mods.value
    if isinstance(mods, Sequence):
        if len(mods) == 0:
            return
        return Mods.parse_any_list(mods).value
    raise TypeError(f"mods argument must be of type Mods or Sequence, not {type(mods)}")


def prettify(cls: object, *fields: str) -> str:
    d = {s: getattr(cls, s) for s in fields}
    return cls.__class__.__qualname__ + '(' + ', '.join([f"{k}={d[k]!r}" for k in d]) + ')'


def parse_enum_args(*args):
    args = [arg.value if isinstance(arg, Enum) else arg for arg in args]
    return args if len(args) != 1 else args[0]


def create_multipart_formdata(data: dict):
    return {k: (None, v, "text/plain", {"charset": "utf-8"}) for k, v in data.items()}


class Util:
    @staticmethod
    def int(value):
        if value is None:
            return
        return int(value)

    @staticmethod
    def float(value):
        if value is None:
            return
        return float(value)


class BeatmapsetSearchFilter:
    """
    Util class that helps for filtering in :func:`Client.search_beatmapsets`.

    Read about each filter under its corresponding function.

    All set functions return the instance of the class to allow for chaining.
    """
    def __init__(self):
        self._filters = {
            "query": None,
            "sort": None,
            "nsfw": None,
            "played": None,
            "r": None,
            "m": None,
            "l": None,
            "g": None,
            "e": None,
            "c": None,
            "s": None,
        }

    def set_query(self, query: str) -> 'BeatmapsetSearchFilter':
        """
        Set the query to search for.

        query: :class:`str`
        """
        self._filters["query"] = query
        return self

    def set_sort(self, sort: Union[BeatmapsetSearchSort, str]) -> 'BeatmapsetSearchFilter':
        """
        Set the sort order of the search.

        sort: Union[:class:`BeatmapsetSearchSort`, :class:`str`]
        """
        self._filters["sort"] = sort.value if isinstance(sort, BeatmapsetSearchSort) else sort
        return self

    def set_nsfw(self, nsfw: bool) -> 'BeatmapsetSearchFilter':
        """
        Set whether to include NSFW in the search.

        nsfw: :class:`bool`
        """
        self._filters["nsfw"] = nsfw
        return self

    def set_played(self, played: Union[BeatmapsetSearchPlayed, str]) -> 'BeatmapsetSearchFilter':
        """
        Set whether to include played and unplayed beatmapsets in the search.

        played: Union[:class:`BeatmapsetSearchPlayed`, :class:`str`]
        """
        self._filters["played"] = played.value if isinstance(played, BeatmapsetSearchPlayed) else played
        return self

    def set_ranked(self, rank: Union[ScoreRank, str]) -> 'BeatmapsetSearchFilter':
        """
        Filter by rank achieved.

        rank: Union[:class:`ScoreRank`, :class:`str`]
        """
        self._filters["r"] = rank.value if isinstance(rank, ScoreRank) else rank
        return self

    def set_mode(self, mode: Union[GameModeInt, str]) -> 'BeatmapsetSearchFilter':
        """
        Set the game mode to filter by.

        mode: Union[:class:`GameModeStr`, :class:`str`]
        """
        self._filters["m"] = mode.value if isinstance(mode, GameModeInt) else mode
        return self

    def set_language(self, language: str) -> 'BeatmapsetSearchFilter':
        """
        Set the language to filter by.

        language: :class:`str`
        """
        self._filters["l"] = language
        return self

    def set_genre(self, genre: str) -> 'BeatmapsetSearchFilter':
        """
        Set the genre to filter by.

        genre: :class:`str`
        """
        self._filters["g"] = genre
        return self

    def set_extra(self, extras: Sequence[Union[BeatmapsetSearchExtra, str]]) -> 'BeatmapsetSearchFilter':
        """
        Set the extras to filter by.

        extras: Sequence[Union[:class:`BeatmapsetSearchExtra`, :class:`str`]]
        """
        self._filters["e"] = ".".join(list(map(
            lambda x: x.value if isinstance(x, BeatmapsetSearchExtra) else x, extras)))
        return self

    def set_generals(self, generals: Sequence[Union[BeatmapsetSearchGeneral, str]]) -> 'BeatmapsetSearchFilter':
        """
        Set the generals to filter by.

        generals: Sequence[Union[:class:`BeatmapsetSearchGeneral`, :class:`str`]]
        """
        self._filters["c"] = ".".join(list(map(
            lambda x: x.value if isinstance(x, BeatmapsetSearchGeneral) else x, generals)))
        return self

    def set_status(self, status: Union[BeatmapsetSearchStatus, str]) -> 'BeatmapsetSearchFilter':
        """
        Set the status to filter by.

        status: Union[:class:`BeatmapsetSearchStatus`, :class:`str`]
        """
        self._filters["s"] = status.value if isinstance(status, BeatmapsetSearchStatus) else status
        return self

    @property
    def filters(self):
        """
        Dictionary of all filters only including ones that have been set.
        """
        return {k: v for k, v in self._filters.items() if v is not None}


class PlaylistItemUtil:
    """
    Util class for passing playlist items to endpoints that involve creating a room
    like create_multiplayer_room and create_playlist.

    **Init Parameters**

    beatmap_id: :class:`int`

    ruleset_id: Union[:class:`GameModeInt`, :class:`int`]

    allowed_mods: Sequence[:class:`Mod`]

    required_mods: Sequence[:class:`Mod`]

    **Properties**

    json: :class:`dict`
        Dictionary format of all the attributes which the client uses
        when sending a http request
    """
    __slots__ = ("beatmap_id", "ruleset_id", "allowed_mods", "required_mods")

    def __init__(self, beatmap_id: int, ruleset_id: Union[GameModeInt, int],
                 allowed_mods: Sequence[Mod] = None, required_mods: Sequence[Mod] = None):
        self.beatmap_id = beatmap_id
        self.ruleset_id = ruleset_id
        self.allowed_mods = allowed_mods
        self.required_mods = required_mods

    @property
    def json(self) -> dict:
        def mod_map(mod):
            return {"acronym": mod.value}
        return {
            "beatmap_id": self.beatmap_id,
            "ruleset_id": parse_enum_args(self.ruleset_id),
            "allowed_mods": list(map(mod_map, self.allowed_mods)) if self.allowed_mods is not None else None,
            "required_mods": list(map(mod_map, self.required_mods)) if self.required_mods is not None else None,
        }


class JsonUtil:
    __slots__ = ()

    def __init__(self, *args):
        for attr, value in zip(self.__slots__, args):
            setattr(self, attr, value)

    def __getitem__(self, item):
        return getattr(self, item)

    def values(self):
        return [getattr(self, attr) for attr in self.__slots__]

    def json(self):
        return {attr: getattr(self, attr) for attr in self.__slots__}

    def items(self):
        return map(lambda attr: (attr, getattr(self, attr)), self.__slots__)

    @staticmethod
    def parse_list(l):
        return list(map(lambda item: item.json if isinstance(item, JsonUtil) else item, l))

    @staticmethod
    def list_to_labeled_dict(l, name):
        return dict(sum(map(
            lambda i: list(map(
                lambda i2: (f"{name}[{i[0]}][{i2[0]}]", f"{i2[1]}"),
                filter(lambda i2: i2[1] is not None, i[1].items())
            )),
            enumerate(l)
        ), []))


class NotificationsUtil(JsonUtil):
    __slots__ = ("category", "id", "object_id", "object_type")

    def __init__(self, category: Union[NotificationCategory, str] = None, id: int = None,
                 object_id: str = None, object_type: Union[ObjectType, str] = None):
        """
        **Arguments**

        category: Optional[:class:`str`]
            Notification category.

        id: Optional[:class:`id`]
            Id of notification to be marked as read

        object_id: Optional[:class:`str`]
            Id of the object that triggered the notification

        object_type: Optional[:class:`str`]
            Type of the object that triggered the notification
        """
        super().__init__(*parse_enum_args(category, id, object_id, object_type))


class IdentitiesUtil(JsonUtil):
    __slots__ = ("category", "object_id", "object_type")

    def __init__(self, category: Union[NotificationCategory, str] = None, object_id: str = None,
                 object_type: Union[ObjectType, str] = None):
        """
        **Arguments**

        category: Optional[:class:`str`]
            Notification category.

        object_id: Optional[:class:`str`]
            Id of the object that triggered the notification

        object_type: Optional[:class:`str`]
            Type of the object that triggered the notification
        """
        super().__init__(*parse_enum_args(category, object_id, object_type))



from .beatmap import BeatmapCompact, BeatmapsetCompact
from .user import UserCompact
from ..enums import GameModeStr, GameModeInt, Mods, Mod, ObjectType
from ..util import prettify
from dateutil import parser


class BeatmapScores:
    """
    Contains a list of scores as well as, possibly, a :class:`BeatmapUserScore` object.

    **Attributes**

    scores: Sequence[Union[:class:`Score`, :class:`SoloScore`]]
        Contains objects of type :class:`Score`. The list of top scores for the beatmap in descending order.

    user_score: :class:`BeatmapUserScore` or :class:`None`
        The score of the current user. This is not returned if the current user does not have a score.
    """
    __slots__ = (
        "scores", "user_score"
    )

    def __init__(self, data, score_type="legacy"):
        self.scores = list(map(LegacyScore if score_type == "legacy" else SoloScore, data['scores']))
        var_name = 'userScore' if 'userScore' in data else 'user_score'
        self.user_score = BeatmapUserScore(data[var_name]) if data.get(var_name) is not None else None

    def __repr__(self):
        return prettify(self, 'user_score', 'scores')


class Score:
    def __new__(cls, data):
        if data["type"] == "solo_score":
            return SoloScore(data)
        return LegacyScore(data)


class LegacyScore:
    """
    Contains information about a score

    **Attributes**

    id: :class:`int`

    best_id: :class:`int`

    user_id: :class:`int`

    accuracy: :class:`float`

    mods: :class:`Mods`

    score: :class:`int`

    max_combo: :class:`int`

    perfect: :class:`bool`

    statistics: :class:`ScoreStatistics`

    passed :class:`bool`

    pp: :class:`float`

    rank: :class:`int`

    created_at: :class:`datetime.datetime`

    mode: :class:`GameModeStr`

    mode_int: :ref:`GameModeInt`

    replay: :class:`bool`
        whether or not the replay is available

    **Optional Attributes**

    beatmap: :class:`BeatmapCompact`

    beatmapset: :class:`BeatmapsetCompact`

    rank_country

    rank_global

    weight

    user

    match
    """
    __slots__ = (
        "id", "best_id", "user_id", "accuracy", "mods", "score", "max_combo", "perfect", "statistics", "passed",
        "pp", "rank", "created_at", "mode", "mode_int", "replay", "beatmap", "beatmapset", "rank_country",
        "rank_global", "weight", "user", "match", "type", "current_user_attributes"
    )

    def __init__(self, data):
        self.id = data['id']
        self.best_id = data['best_id']
        self.user_id = data['user_id']
        self.accuracy = data['accuracy']
        self.mods = Mods.parse_any_list(data['mods'])
        self.score = data['score']
        self.max_combo = data['max_combo']
        self.perfect = data['perfect']
        self.statistics = ScoreStatistics(data['statistics'])
        self.passed = data['passed']
        self.pp = data['pp']
        self.rank = data['rank']
        self.created_at = parser.parse(data['created_at'])
        self.mode = GameModeStr(data['mode'])
        self.mode_int = GameModeInt(data['mode_int'])
        self.replay = data['replay']
        self.type = ObjectType(data['type'])

        self.beatmap = BeatmapCompact(data['beatmap']) if 'beatmap' in data else None
        self.beatmapset = BeatmapsetCompact(data['beatmapset']) if 'beatmapset' in data else None
        self.user = UserCompact(data['user']) if 'user' in data else None
        self.match = data['match'] if 'match' in data else None
        self.rank_country = data['rank_country'] if 'rank_country' in data else None
        self.rank_global = data['rank_global'] if 'rank_global' in data else None
        self.weight = data['weight'] if 'weight' in data else None

    def __repr__(self):
        return prettify(self, 'user_id', 'accuracy')


class SoloScore:
    """
    Contains information about a lazer score.

    **Attributes**

    accuracy: :class:`float`

    beatmap_id: :class:`int`

    build_id: :class:`int`

    ended_at: :class:`datetime.datetime`

    legacy_score_id: :class:`int`

    legacy_total_score: :class:`int`

    max_combo: :class:`int`

    maximum_statistics: :class:`ScoreDataStatistics`

    mods: Sequence[:class:`LazerMod`]

    passed: :class:`bool`

    rank: :class:`str`

    ruleset_id: :class:`int`

    started_at: :class:`datetime.datetime`

    statistics: :class:`ScoreDataStatistics`

    total_score: :class:`int`

    user_id: :class:`int`

    best_id: :class:`int`

    id: :class:`int`

    legacy_perfect: :class:`bool`

    pp: :class:`float`

    replay: :class:`bool`

    type: :class:`ObjectType`

    user: :class:`UserCompact`
    """
    __slots__ = (
        "accuracy", "beatmap_id", "build_id", "ended_at", "legacy_score_id",
        "legacy_total_score", "max_combo", "maximum_statistics", "mods",
        "passed", "rank", "ruleset_id", "started_at", "statistics", "total_score",
        "user_id", "best_id", "id", "legacy_perfect", "pp", "replay", "type", "user"
    )

    def __init__(self, data):
        self.accuracy = data["accuracy"]
        self.beatmap_id = data["beatmap_id"]
        self.ended_at = parser.parse(data["ended_at"])
        self.max_combo = data["max_combo"]
        self.maximum_statistics = ScoreDataStatistics(data["maximum_statistics"])
        self.mods = list(map(LazerMod, data["mods"]))
        self.passed = data["passed"]
        self.rank = data["rank"]
        self.ruleset_id = data["ruleset_id"]
        self.statistics = ScoreDataStatistics(data["statistics"])
        self.total_score = data["total_score"]
        self.user_id = data["user_id"]
        self.best_id = data["best_id"]
        self.id = data["id"]
        self.legacy_perfect = data["legacy_perfect"]
        self.pp = data["pp"]
        self.replay = data["replay"]
        self.type = ObjectType(data["type"])
        self.user = UserCompact(data['user'])

        self.build_id = data.get("build_id")
        self.legacy_score_id = data.get("legacy_score_id")
        self.legacy_total_score = data.get("legacy_total_score")
        self.started_at = parser.parse(data["started_at"]) if data.get("started_at") is not None else None

    def __repr__(self):
        return prettify(self, "beatmap_id", "statistics")


class ScoreStatistics:
    """
    **Attributes**

    count_50: :class:`int`

    count_100: :class:`int`

    count_300: :class:`int`

    count_geki: :class:`int`

    count_katu: :class:`int`

    count_miss: :class:`int`
    """
    __slots__ = (
        "count_50", "count_100", "count_300", "count_geki",
        "count_katu", "count_miss"
    )

    def __init__(self, data):
        self.count_50 = data['count_50']
        self.count_100 = data['count_100']
        self.count_300 = data['count_300']
        self.count_geki = data['count_geki']
        self.count_katu = data['count_katu']
        self.count_miss = data['count_miss']

    def __repr__(self):
        return prettify(self, 'count_300', 'count_miss')


class ScoreDataStatistics:
    """
    **Attributes**

    ok: Union[:class:`int`, :class:`None`]

    meh: Union[:class:`int`, :class:`None`]

    good: Union[:class:`int`, :class:`None`]

    miss: Union[:class:`int`, :class:`None`]

    none: Union[:class:`int`, :class:`None`]

    great: Union[:class:`int`, :class:`None`]

    perfect: Union[:class:`int`, :class:`None`]

    ignore_hit: Union[:class:`int`, :class:`None`]

    ignore_miss: Union[:class:`int`, :class:`None`]

    large_bonus: Union[:class:`int`, :class:`None`]

    small_bonus: Union[:class:`int`, :class:`None`]

    large_tick_hit: Union[:class:`int`, :class:`None`]

    small_tick_hit: Union[:class:`int`, :class:`None`]

    large_tick_miss: Union[:class:`int`, :class:`None`]

    small_tick_miss: Union[:class:`int`, :class:`None`]
    """
    __slots__ = (
        'ok', "meh", 'good', 'miss', 'none', 'great', 'perfect', 'ignore_hit',
        'ignore_miss', 'large_bonus', 'small_bonus', 'large_tick_hit', 'small_tick_hit',
        'large_tick_miss', 'small_tick_miss'
    )

    def __init__(self, data):
        self.ok = data.get('ok')
        self.meh = data.get('meh')
        self.good = data.get('good')
        self.miss = data.get('miss')
        self.none = data.get('none')
        self.great = data.get('great')
        self.perfect = data.get('perfect')
        self.ignore_hit = data.get('ignore_hit')
        self.ignore_miss = data.get('ignore_miss')
        self.large_bonus = data.get('large_bonus')
        self.small_bonus = data.get('small_bonus')
        self.large_tick_hit = data.get('large_tick_hit')
        self.small_tick_hit = data.get('small_tick_hit')
        self.large_tick_miss = data.get('large_tick_miss')
        self.small_tick_miss = data.get('small_tick_miss')

    def __repr__(self):
        return prettify(self, 'perfect', 'miss')


class LazerMod:
    """
    **Attributes**

    mod: Union[:class:`Mod`, :class:`str`]
        :class:`str` if mod acronym doesn't exist in :class:`Mod` enum

    settings: :class:`dict` or :class:`None`
    """
    __slots__ = ("mod", "settings")

    def __init__(self, data):
        try:
            self.mod = Mod(data['acronym'])
        except ValueError:
            self.mod = data['acronym']
        self.settings = data.get('settings')

    def __repr__(self):
        return prettify(self, 'mod', 'settings')


class BeatmapUserScore:
    """
    **Attributes**

    position: :class:`int`
        The position of the score within the requested beatmap ranking.

    score: :class:`LegacyScore`
        The details of the score.
    """
    __slots__ = (
        "position", "score"
    )

    def __init__(self, data):
        self.position = data.get('position')
        self.score = LegacyScore(data['score'])

    def __repr__(self):
        return prettify(self, 'position')



from .user import UserCompact
from dateutil import parser

from ..util import prettify


class SeasonalBackgrounds:
    """
    Contains data on the seasonal backgrounds.

    **Attributes**

    ends_at: :class:`datetime.datetime`
        The date when the seasonal backgrounds will end.

    backgrounds: Sequence[:class:`SeasonalBackground`]
        A list of all the seasonal backgrounds.
    """
    __slots__ = ('ends_at', 'backgrounds')

    def __init__(self, data):
        self.ends_at = parser.parse(data['ends_at'])
        self.backgrounds = list(map(SeasonalBackground, data['backgrounds']))

    def __repr__(self):
        return prettify(self, 'ends_at', 'backgrounds')


class SeasonalBackground:
    """
    Represents a seasonal background.

    **Attributes**

    url: :class:`str`
        The url of the background.

    user: :class:`UserCompact`
        The artist of the background.
    """
    __slots__ = ('url', 'user')

    def __init__(self, data):
        self.url = data['url']
        self.user = UserCompact(data['user'])

    def __repr__(self):
        return prettify(self, 'url', 'user')



from .group import UserGroup
from ..util import Util, prettify
from ..enums import GameModeStr
from dateutil import parser
import math


class UserCompact:
    """
    Mainly used for embedding in certain responses to save additional api lookups.

    **Attributes**

    avatar_url: :class:`str`
        url of user's avatar

    country_code: :class:`str`
        two-letter code representing user's country

    default_group: :class:`str`
        Identifier of the default Group the user belongs to.

    id: :class:`int`
        unique identifier for user

    is_active: :class:`bool`
        has this account been active in the last x months?

    is_bot: :class:`bool`
        is this a bot account?

    is_deleted: :class:`bool`

    is_online: :class:`bool`
        is the user currently online? (either on lazer or the new website)

    is_supporter: :class:`bool`
        does this user have supporter?

    last_visit: :class:`datetime.datetime`
        null if the user hides online presence

    pm_friends_only: :class:`bool`
        whether or not the user allows PM from other than friends

    profile_colour: :class:`str`
        colour of username/profile highlight, hex code (e.g. #333333)

    username: :class:`str`
        user's display name

    **Possible Attributes**

    account_history: Sequence[:class:`UserAccountHistory`]

    active_tournament_banner: :class:`ProfileBanner`

    badges: Sequence[:class:`UserBadge`]

    beatmap_playcounts_count: :class:`int`

    blocks

    country

    cover

    favourite_beatmapset_count: :class:`int`

    follower_count: :class:`int`

    mapping_follower_count: :class:`int`

    friends

    graveyard_beatmapset_count: :class:`int`

    groups: Sequencep[:class:`UserGroup`]

    is_restricted: :class:`bool`

    loved_beatmapset_count: :class:`int`

    monthly_playcounts: Sequence[:class:`UserMonthlyPlaycount`]

    page

    previous_usernames

    ranked_beatmapset_count

    replays_watched_counts

    scores_best_count: :class:`int`

    scores_first_count: :class:`int`

    scores_recent_count: :class:`int`

    statistics

    statistics_rulesets: :class:`UserStatisticsRulesets`

    support_level

    pending_beatmapset_count

    unread_pm_count

    user_achievements

    user_preferences

    rank_history
    """
    __slots__ = (
        'avatar_url', 'country_code', 'default_group', 'id', 'is_active', 'is_bot', 'is_deleted', 'is_online',
        'is_supporter', 'last_visit', 'pm_friends_only', 'profile_colour', 'username', 'account_history',
        'active_tournament_banner', 'badges', 'beatmap_playcounts_count', 'blocks', 'country', 'cover',
        'favourite_beatmapset_count', 'follower_count', 'mapping_follower_count', 'friends',
        'graveyard_beatmapset_count', 'groups', 'is_restricted', 'loved_beatmapset_count', 'monthly_playcounts',
        'page', 'previous_usernames', 'ranked_beatmapset_count', 'replays_watched_counts', 'scores_best_count',
        'scores_first_count', 'scores_recent_count', 'statistics', 'statistics_rulesets', 'support_level',
        'pending_beatmapset_count', 'unread_pm_count', 'user_achievements', 'user_preferences', 'rank_history',
        'ranked_beatmapset_counts'
    )

    def __init__(self, data):
        self.avatar_url = data['avatar_url']
        self.country_code = data['country_code']
        self.default_group = data['default_group']
        self.id = data['id']
        self.is_active = data['is_active']
        self.is_bot = data['is_bot']
        self.is_deleted = data['is_deleted']
        self.is_online = data['is_online']
        self.is_supporter = data['is_supporter']
        self.last_visit = parser.parse(data['last_visit']) if data['last_visit'] is not None else None
        self.pm_friends_only = data['pm_friends_only']
        self.profile_colour = data['profile_colour']
        self.username = data['username']

        # Optional attributes
        self.active_tournament_banner = ProfileBanner(data['active_tournament_banner']) \
            if data.get('active_tournament_banner') is not None else None
        self.account_history = list(map(UserAccountHistory, data.get('account_history', [])))
        self.badges = list(map(UserBadge, data.get('badges', [])))
        self.groups = list(map(UserGroup, data.get('groups', [])))
        self.monthly_playcounts = list(map(UserMonthlyPlaycount, data.get('monthly_playcounts', []))) \
            if data.get("monthly_playcounts") is not None else None
        self.statistics = UserStatistics(data['statistics']) if 'statistics' in data else None
        self.page = data.get('page')
        self.pending_beatmapset_count = Util.int(data.get('pending_beatmapset_count'))
        self.previous_usernames = data.get('previous_usernames')
        self.rank_history = data.get('rank_history')
        self.ranked_beatmapset_counts = data.get('ranked_beatmapset_counts')
        self.replays_watched_counts = data.get('replays_watched_counts')
        self.scores_best_count = Util.int(data.get('scores_best_count'))
        self.scores_first_count = Util.int(data.get('scores_first_count'))
        self.scores_recent_count = Util.int(data.get('scores_recent_count'))
        self.statistics_rulesets = UserStatisticsRulesets(data['statistics_rulesets']) \
            if 'statistics_rulesets' in data else None  # TODO
        self.support_level = data.get('support_level')
        self.unread_pm_count = Util.int(data.get('unread_pm_count'))
        self.user_achievements = data.get('user_achievements')
        self.user_preferences = data.get('user_preferences')
        self.beatmap_playcounts_count = Util.int(data.get('beatmap_playcounts_count'))
        self.blocks = data.get('blocks')
        self.country = data.get('country')
        self.cover = data.get('cover')
        self.favourite_beatmapset_count = Util.int(data.get('favourite_beatmapset_count'))
        self.follower_count = Util.int(data.get('follower_count'))
        self.mapping_follower_count = Util.int(data.get('mapping_follower_count'))
        self.friends = data.get('friends')
        self.graveyard_beatmapset_count = Util.int(data.get('graveyard_beatmapset_count'))
        self.is_restricted = data.get('is_restricted')
        self.loved_beatmapset_count = Util.int(data.get('loved_beatmapset_count'))

    def __repr__(self):
        return prettify(self, 'username', 'id')


class User(UserCompact):
    """
    Represents a User. Extends UserCompact object with additional attributes.

    **Attributes**

    cover_url: :class:`str`
        url of profile cover. Deprecated, use cover['url'] instead.

    discord: :class:`str` or :class:`NoneType`

    has_supported: :class:`bool`
        whether or not ever being a supporter in the past

    interests: :class:`str` or :class:`NoneType`

    join_date: :class:`datetime.datetime`

    kudosu: :class:`dict`
        Contains items available: :class:`int` and total: :class:`int`

    location: :class:`str` or :class:`NoneType`

    max_blocks: :class:`int`
        maximum number of users allowed to be blocked

    max_friends: :class:`int`
        maximum number of friends allowed to be added

    occupation: :class:`str` or :class:`NoneType`

    playmode: :class:`GameModeStr`

    playstyle: Sequence[:class:`str`]
        Device choices of the user.

    post_count: :class:`int`
        number of forum posts

    profile_order: Sequence[:class:`str`]
        Ordered list of sections in user profile page. Sections consist of:
        me, recent_activity, beatmaps, historical, kudosu, top_ranks, medals

    title: :class:`str` or :class:`NoneType`
        user-specific title

    title_url: :class:`str` or :class:`NoneType`

    twitter: :class:`str` or :class:`NoneType`

    website: :class:`str` or :class:`NoneType`

    country: :class:`dict`
        Contains items code: :class:`str` and name: :class:`str`

    cover: :class:`dict`
        Contains items custom_url: :class:`str`, url: :class:`str`, and id: :class:`int`.

    is_restricted: :class:`bool`
        present only if this is the currently authenticated user

    **Possible Attributes**

    All possible attributes come from :class:`UserCompact`
    """
    __slots__ = (
        'cover_url', 'discord', 'has_supported', 'interests', 'join_date', 'kudosu', 'location', 'max_blocks',
        'max_friends', 'occupation', 'playmode', 'playstyle', 'post_count', 'profile_order', 'title', 'title_url',
        'twitter', 'website'
    )

    def __init__(self, data):
        super().__init__(data)
        self.cover_url = data['cover']
        self.discord = data['discord']
        self.has_supported = data['has_supported']
        self.interests = data['interests']
        self.join_date = parser.parse(data['join_date'])
        self.kudosu = data['kudosu']
        self.location = data['location']
        self.max_blocks = data['max_blocks']
        self.max_friends = data['max_friends']
        self.occupation = data['occupation']
        self.playmode = GameModeStr(data['playmode'])
        self.playstyle = data['playstyle']
        self.post_count = data['post_count']
        self.profile_order = data['profile_order']
        self.title = data['title']
        self.title_url = data['title_url']
        self.twitter = data['twitter']
        self.website = data['website']

    def __str__(self):
        return super().__repr__()


class ProfileBanner:
    """
    **Attributes**

    id: :class:`int`

    tournament_id: :class:`int`

    image: :class:`str`

    image2x: :class:`str`
    """
    __slots__ = (
        "id", "tournament_id", "image", "image2x"
    )

    def __init__(self, data):
        print(data)
        self.id = data['id']
        self.tournament_id = data['tournament_id']
        self.image = data['image']
        self.image2x = data['image@2x']

    def __repr__(self):
        return prettify(self, 'tournament_id')


class UserSilence:
    """
    **Attributes**

    id: :class:`int`
        id of this object.

    user_id: :class:`int`
        id of the User that was silenced
    """
    __slots__ = (
        "id", "user_id"
    )

    def __init__(self, data):
        self.id = data['id']
        self.user_id = data['user_id']

    def __repr__(self):
        return prettify(self, 'user_id')


class UserAccountHistory:
    """
    **Attributes**

    id: :class:`int`

    type: :class:`str`
        Can be one of the following: note, restriction, or silence.

    timestamp: :class:`datetime.datetime`

    length: :class:`int`
        In seconds.
    """
    __slots__ = (
        "id", "type", "timestamp", "length"
    )

    def __init__(self, data):
        self.id = data['id']
        self.type = data['type']
        self.timestamp = parser.parse(data['timestamp'])
        self.length = data['length']

    def __repr__(self):
        return prettify(self, 'type', 'length')


class UserBadge:
    """
    **Attributes**

    awarded_at: :class:`datetime.datetime`

    description: :class:`str`

    image_url: :class:`str`

    url: :class:`str`
    """
    __slots__ = (
        "awarded_at", "description", "image_url", "url"
    )

    def __init__(self, data):
        self.awarded_at = parser.parse(data['awarded_at'])
        self.description = data['description']
        self.image_url = data['image_url']
        self.url = data['url']

    def __repr__(self):
        return prettify(self, 'awarded_at')


class UserMonthlyPlaycount:
    """
    **Attributes**

    start_date: :class:`str`
        year-month-day format

    count: class:`int`
        playcount
    """
    __slots__ = (
        "start_date", "count"
    )

    def __init__(self, data):
        self.start_date = data['start_date']
        self.count = data['count']

    def __repr__(self):
        return prettify(self, 'start_date', 'count')


class UserStatistics:
    """
    A summary of various gameplay statistics for a User. Specific to a :class:`GameMode`

    **Attributes**

    count_300: :class:`int`

    count_100: :class:`int`

    count_50: :class:`int`

    count_miss: :class:`int`

    grade_counts: :class:`dict`
        Below are the keys, their type, and meaning.

        a: :class:`int`
            Number of A ranked scores.

        s: :class:`int`
            Number of S ranked scores.

        sh: :class:`int`
            Number of Silver S ranked scores.

        ss: :class:`int`
            Number of SS ranked scores.

        ssh: :class:`int`
            Number of Silver SS ranked scores.

    hit_accuracy: :class:`int`
        Hit accuracy percentage

    is_ranked: :class:`bool`
        Is actively ranked

    level: :class:`dict`
        Contains keys 'current' (current level) and 'progress' (progress to next level).

    maximum_combo: :class:`int`
        Highest maximum combo.

    play_count: :class:`int`
        Number of maps played.

    play_time: :class:`int`
        Cumulative time played.

    pp: :class:`int`
        Performance points

    pp_exp: :class:`int`
        Experimental performance points (on lazer.ppy.sh)

    recommended_difficulty: :class:`float`
        Recommended difficulty for a player. This value is not received from the api, but locally calculated.

    global_rank: Union[:class:`int`, :class:`None`]
        Current global rank according to pp.

    global_rank_exp: Union[:class:`int`, :class:`None`]
        Current global rank according to experimental pp.

    country_rank: Union[:class:`int`, :class:`None`]
        Current country rank according to pp.

    ranked_score: :class:`int`
        Current ranked score.

    replays_watched_by_others: :class:`int`
        Number of replays watched by other users.

    total_hits: :class:`int`
        Total number of hits.

    total_score: :class:`int`
        Total score.

    user: Union[:class:`UserCompact`, :class:`None`]
        The associated user.
    """
    __slots__ = (
        "grade_counts", "level", "hit_accuracy", "is_ranked", "maximum_combo",
        "play_count", "play_time", "pp", "global_rank", "ranked_score",
        "replays_watched_by_others", "total_hits", "total_score", "user",
        "country_rank", "global_rank_exp", "pp_exp", "count_100", "count_300",
        "count_50", "count_miss"
    )

    def __init__(self, data):
        self.count_100 = data['count_100']
        self.count_300 = data['count_300']
        self.count_50 = data['count_50']
        self.count_miss = data['count_miss']
        self.grade_counts = data['grade_counts']
        self.level = data['level']
        self.hit_accuracy = data['hit_accuracy']
        self.is_ranked = data['is_ranked']
        self.maximum_combo = data['maximum_combo']
        self.play_count = data['play_count']
        self.play_time = data['play_time']
        self.pp = data['pp']
        self.pp_exp = data.get('pp_exp')
        self.global_rank = data['global_rank']
        self.global_rank_exp = data.get('global_rank_exp')
        self.country_rank = data.get('country_rank')
        self.ranked_score = data['ranked_score']
        self.replays_watched_by_others = data['replays_watched_by_others']
        self.total_hits = data['total_hits']
        self.total_score = data['total_score']
        self.user = UserCompact(data['user']) if data.get("user") is not None else None

    @property
    def recommended_difficulty(self):
        return math.pow(self.pp, 0.4) * 0.195

    def __repr__(self):
        return prettify(self, 'pp', 'global_rank', 'user')


class UserStatisticsRulesets:
    """
    Object that contains statistics for each gamemode.

    **Attributes**

    osu: :class:`UserStatistics`
        statistics for osu!standard.

    taiko: :class:`UserStatistics`
        statistics for osu!taiko.

    fruits: :class:`UserStatistics`
        statistics for osu!catch.

    mania: :class:`UserStatistics`
        statistics for osu!mania.
    """
    __slots__ = ('osu', 'taiko', 'fruits', 'mania')

    def __init__(self, data):
        self.osu = UserStatistics(data['osu']) if data.get('osu') else None
        self.taiko = UserStatistics(data['taiko']) if data.get('taiko') else None
        self.fruits = UserStatistics(data['fruits']) if data.get('fruits') else None
        self.mania = UserStatistics(data['mania']) if data.get('mania') else None

    def __repr__(self):
        fields = [getattr(self, slot) for slot in self.__slots__ if getattr(self, slot, None)]
        return prettify(self, *fields)


class CurrentUserAttributes:
    """
    An object listing various related permissions and states for the current user,
    related to the object it is attached to.

    **BeatmapDiscussionPermissions Attributes**

    can_destroy: :class:`bool`
        Can delete the discussion.

    can_reopen: :class:`bool`
        Can reopen the discussion.

    can_moderate_kudosu: :class:`bool`
        Can allow or deny kudosu.

    can_resolve: :class:`bool`
        Can resolve the discussion.

    vote_score: :class:`bool`
        Current vote given to the discussion.

    **ChatChannelUserAttributes Attributes**

    can_message: :class:`bool`
        Can send messages to this channel.

    can_message_error: :class:`str`
        Reason messages cannot be sent to this channel

    last_read_id: :class:`int`
        message_id of last message read.
    """
    def __init__(self, data, attr_type):
        self.type = attr_type
        if attr_type == "BeatmapsetDiscussionPermissions":
            self.can_destroy = data['can_destroy']
            self.can_reopen = data['can_reopen']
            self.can_moderate_kudosu = data['can_moderate_kudosu']
            self.can_resolve = data['can_resolve']
            self.vote_score = data['vote_score']
        elif attr_type == "ChatChannelUserAttributes":
            self.can_message = data['can_message']
            self.can_message_error = data['can_message_error']
            self.last_read_id = data['last_read_id']
        else:
            print(f"WARNING: Unrecognized attr_type for CurrentUserAttributes: \"{attr_type}\"")
            for k, v in data.items():
                setattr(self, k, v)

    def __repr__(self):
        return prettify(self, 'type')
